"""Functions that can be used to test multi-objective algorithms."""
from math import inf, isfinite
from typing import Any, Final

import numpy as np
from numpy import array_equal
from numpy.random import Generator, default_rng

from moptipy.api.algorithm import (
    Algorithm0,
    Algorithm1,
    Algorithm2,
    check_algorithm,
)
from moptipy.api.encoding import Encoding
from moptipy.api.mo_algorithm import MOAlgorithm
from moptipy.api.mo_archive import MOArchivePruner
from moptipy.api.mo_execution import MOExecution
from moptipy.api.mo_problem import MOProblem
from moptipy.api.operators import check_op0, check_op1, check_op2
from moptipy.api.space import Space
from moptipy.mo.archive.keep_farthest import KeepFarthest
from moptipy.tests.component import validate_component
from moptipy.tests.encoding import validate_encoding
from moptipy.tests.mo_problem import validate_mo_problem
from moptipy.tests.space import validate_space
from moptipy.utils.nputils import rand_seed_generate
from moptipy.utils.types import check_int_range, type_error


def validate_mo_algorithm(
        algorithm: MOAlgorithm,
        solution_space: Space,
        problem: MOProblem,
        search_space: Space | None = None,
        encoding: Encoding | None = None,
        max_fes: int = 100,
        is_encoding_deterministic: bool = True) -> None:
    """
    Check whether a multi-objective algorithm follows the moptipy API.

    :param algorithm: the algorithm to test
    :param solution_space: the solution space
    :param problem: the problem to solve
    :param search_space: the optional search space
    :param encoding: the optional encoding
    :param max_fes: the maximum number of FEs
    :param is_encoding_deterministic: is the encoding deterministic?
    :raises TypeError: if `algorithm` is not a
        :class:`~moptipy.api.mo_algorithm.MOAlgorithm` instance
    :raises ValueError: if `algorithm` does not behave like it should
    """
    if not isinstance(algorithm, MOAlgorithm):
        raise type_error(algorithm, "algorithm", MOAlgorithm)

    check_algorithm(algorithm)
    if isinstance(algorithm, Algorithm0):
        check_op0(algorithm.op0)
        if isinstance(algorithm, Algorithm1):
            check_op1(algorithm.op1)
            if isinstance(algorithm, Algorithm2):
                check_op2(algorithm.op2)

    validate_component(algorithm)
    validate_mo_problem(problem, None, None)
    validate_space(solution_space, None)

    if encoding is not None:
        validate_encoding(encoding, None, None, None,
                          is_encoding_deterministic)
        validate_space(search_space, None)

    check_int_range(max_fes, "max_fes", 1, 1_000_000_000)
    lb: Final[int | float] = problem.lower_bound()
    if (not isfinite(lb)) and (lb != -inf):
        raise ValueError(f"objective lower bound cannot be {lb}.")
    ub = problem.upper_bound()
    if (not isfinite(ub)) and (ub != inf):
        raise ValueError(f"objective upper bound cannot be {ub}.")

    exp = MOExecution()
    exp.set_algorithm(algorithm)
    exp.set_max_fes(max_fes)
    exp.set_solution_space(solution_space)
    exp.set_objective(problem)
    if search_space is not None:
        exp.set_search_space(search_space)
        exp.set_encoding(encoding)

    random: Final[Generator] = default_rng()
    max_archive_size: Final[int] = int(random.integers(
        1, 1 << int(random.integers(1, 6))))
    exp.set_archive_max_size(max_archive_size)
    exp.set_archive_pruning_limit(
        max_archive_size + int(random.integers(0, 8)))
    if random.integers(2) <= 0:
        choice: int = int(random.integers(2))
        pruner: MOArchivePruner
        if choice <= 0:
            lst: list[int]
            while True:
                lst = [i for i in range(problem.f_dimension())
                       if random.integers(2) <= 0]
                if len(lst) > 0:
                    break
            pruner = KeepFarthest(problem, lst)
        else:
            pruner = MOArchivePruner()
        exp.set_archive_pruner(pruner)

    seed: Final[int] = rand_seed_generate(random)
    exp.set_rand_seed(seed)

    l_consumed_fes: int = -1
    l_last_improvement_fe: int = -1
    l_res_f: int | float = inf
    l_y: Any = None
    l_fs: np.ndarray | None = None
    l_x: Any = None
    for is_check in [False, True]:
        with exp.execute() as process:
            # re-raise any exception that was caught
            if hasattr(process, "_caught"):
                error = getattr(process, "_caught")
                if error is not None:
                    raise error
            # no exception? ok, let's check the data
            if not process.has_best():
                raise ValueError(
                    "The algorithm did not produce any solution.")

            if not process.should_terminate():
                raise ValueError(
                    "The algorithm stopped before hitting the "
                    "termination criterion.")
            consumed_fes: int = check_int_range(
                process.get_consumed_fes(), "consumed_fes", 1, max_fes)
            if is_check:
                if consumed_fes != l_consumed_fes:
                    raise ValueError(
                        f"consumed FEs changed from {l_consumed_fes} to "
                        f"{consumed_fes} in second run for seed {seed}")
            else:
                l_consumed_fes = consumed_fes
            last_improvement_fe = process.get_last_improvement_fe()
            check_int_range(last_improvement_fe, "last_improvement_fe",
                            1, consumed_fes)
            if is_check:
                if last_improvement_fe != l_last_improvement_fe:
                    raise ValueError(
                        "last improvement FEs changed from "
                        f"{l_last_improvement_fe} to {last_improvement_fe} in"
                        f" second run for seed {seed}")
            else:
                l_last_improvement_fe = last_improvement_fe

            consumed_time: int = check_int_range(
                process.get_consumed_time_millis(), "consumed_time",
                0, 100_0000_000)
            check_int_range(
                process.get_last_improvement_time_millis(),
                "last_improvement_time", 0, consumed_time)

            if lb != process.lower_bound():
                raise ValueError(
                    "Inconsistent lower bounds between process "
                    f"({process.lower_bound()}) and scalarized "
                    f"objective ({lb}).")
            if ub != process.upper_bound():
                raise ValueError(
                    "Inconsistent upper bounds between process "
                    f"({process.upper_bound()}) and scalarized "
                    f"objective ({ub}).")

            res_f: float | int = process.get_best_f()
            if not isfinite(res_f):
                raise ValueError(
                    "Infinite scalarized objective value of result.")
            if (res_f < lb) or (res_f > ub):
                raise ValueError(
                    f"Objective value {res_f} outside of bounds [{lb},{ub}].")
            if is_check:
                if res_f != l_res_f:
                    raise ValueError(
                        f"result f changed from {l_res_f} to {res_f} in"
                        f" second run for seed {seed}")
            else:
                l_res_f = res_f

            y = solution_space.create()
            process.get_copy_of_best_y(y)
            solution_space.validate(y)
            fs1 = problem.f_create()
            fs2 = problem.f_create()
            process.get_copy_of_best_y(y)
            check_f = problem.f_evaluate(y, fs1)
            if check_f != res_f:
                raise ValueError(
                    f"Inconsistent objective value {res_f} from process "
                    f"compared to {check_f} from objective function.")
            process.get_copy_of_best_fs(fs2)
            if not array_equal(fs1, fs2):
                raise ValueError(
                    f"Inconsistent objective vectors {fs1} and {fs2}.")

            if is_check:
                if not solution_space.is_equal(y, l_y):
                    raise ValueError(f"solution changed from {l_y} to {y} in "
                                     f"the second run of seed {seed}")
                if res_f != l_res_f:
                    raise ValueError(
                        f"result f changed from {l_res_f} to {res_f} in the "
                        f"second run of seed {seed}")
                if not np.array_equal(fs1, l_fs):
                    raise ValueError(
                        f"result fs changed from {l_fs} to {fs1} in the "
                        f"second run of seed {seed}")
            else:
                l_y = y
                l_fs = fs1
                l_res_f = res_f

            x: Any | None = None
            if search_space is not None:
                x = search_space.create()
                process.get_copy_of_best_x(x)
                search_space.validate(x)

                if is_check:
                    if not search_space.is_equal(x, l_x):
                        raise ValueError(
                            f"result x changed from {l_x} to {x} in the "
                            f"second run of seed {seed}")
                else:
                    l_x = x

            if encoding is not None:
                y2 = solution_space.create()
                encoding.decode(x, y2)
                solution_space.validate(y2)
                if is_encoding_deterministic:
                    solution_space.is_equal(y, y2)



"""
Selection algorithms are common modules that choose `n` out of `N` objects.

Selection algorithms are modules of the fully-configurable Evolutionary
Algorithm :class:`~moptipy.algorithms.so.general_ea.GeneralEA`. They can
utilize fitness values computed by the fitness assignment processes
(:class:`~moptipy.algorithms.so.fitness.Fitness`). Of course, they can
also be applied in different contexts and are not bound to single-objective
optimization.

:class:`~moptipy.algorithms.modules.selection.Selection` is especially
important in Evolutionary Algorithms
(`~moptipy.algorithms.so.general_ea.GeneralEA`), where it is used in two
places: As *survival selection*, it chooses which points will be allowed to
remain in the population and, hence, survive into the mating pool for the next
generation. As *mating selection* methods, they choose the inputs of the
search operations from the mating pool.

:class:`~moptipy.algorithms.modules.selection.Selection` algorithms must
*only* use the
:attr:`~moptipy.algorithms.modules.selection.FitnessRecord.fitness` of a
solution record (and random numbers) to make their decisions. These fitness
values are subject to minimization. They can equal to the objective values in
optimization or stem from a :class:`~moptipy.algorithms.so.fitness.Fitness`
Assignment Process.

The following selection algorithms have currently been implemented:

- :class:`~moptipy.algorithms.modules.selections.best.Best` selection
  selects the best `n` solutions without replacement. This is a common
  strategy for survival selection, especially in (mu+lambda) EAs
  (compatible to :class:`~moptipy.algorithms.so.ea.EA`).
- :class:`~moptipy.algorithms.modules.selections.random_without_repl\
.RandomWithoutReplacement` selects random solutions without replacement. It is
  a common strategy for mating selection.
- :class:`~moptipy.algorithms.modules.selections.fitness_proportionate_sus\
.FitnessProportionateSUS` performs fitness proportionate selection for
  minimization using stochastic uniform sampling and, optionally, a minimum
  selection probability threshold. It is the classic survival selection
  algorithm in Genetic Algorithm.
- :class:`~moptipy.algorithms.modules.selections.tournament_with_repl.\
TournamentWithReplacement`
  performs tournament selection with a specified tournament size with
  replacement.
- :class:`~moptipy.algorithms.modules.selections.tournament_without_repl.\
TournamentWithoutReplacement`
  performs tournament selection with a specified tournament size without
  replacement.
"""
from typing import Any, Callable, Protocol

from numpy.random import Generator

from moptipy.api.component import Component
from moptipy.utils.types import type_error


# start book
class FitnessRecord(Protocol):
    """A fitness record stores data together with a fitness."""

# end book
    #: the fitness value, the only criterion to be used by a selection
    #: algorithm (besides random numbers)
    fitness: int | float

    def __lt__(self, other) -> bool:
        """
        Compare the fitness of this record with the fitness of another one.

        :param other: the other fitness record
        """
        return self.fitness < other.fitness
# start book


class Selection(Component):
    """The base class for selections algorithms."""

    def select(self, source: list[FitnessRecord],
               dest: Callable[[FitnessRecord], Any],
               n: int, random: Generator) -> None:  # pylint: disable=W0613
        """
        Select `n` records from `source` and pass them to `dest`.

        When choosing the `n` records from `source` to be passed to
        `dest`, only the :attr:`~FitnessRecord.fitness` attribute of
        the records and the random numbers from `random` must be used
        as decision criteria.

        :param source: the list with the records to select from
        :param dest: the destination collector Callable to invoke for
            each selected record, can be :class:`list`.`append`.
        :param n: the number of records to select
        :param random: the random number generator
        """
# end book


def check_selection(selection: Selection) -> Selection:
    """
    Check whether an object is a valid instance of :class:`Selection`.

    :param selection: the Selection object
    :return: the object
    :raises TypeError: if `selections` is not an instance of
        :class:`Selection` or if it is an instance of the abstract base
        class
    """
    if not isinstance(selection, Selection):
        raise type_error(selection, "selections", Selection)
    if selection.__class__ is Selection:
        raise TypeError("cannot use abstract class 'Selection' directly")
    return selection



"""Code for selecting interesting instances for smaller-scale experiments."""
import multiprocessing as mp
from math import ceil, factorial, inf, log2, sqrt
from os import sched_getaffinity
from typing import Callable, Final

import numpy as np  # type: ignore
from numpy.random import Generator, RandomState  # type: ignore
from sklearn.cluster import SpectralClustering  # type: ignore

from moptipy.algorithms.so.rls import RLS
from moptipy.api.execution import Execution
from moptipy.examples.jssp.gantt_space import GanttSpace
from moptipy.examples.jssp.instance import Instance
from moptipy.examples.jssp.makespan import Makespan
from moptipy.examples.jssp.ob_encoding import OperationBasedEncoding
from moptipy.operators.permutations.op0_shuffle import Op0Shuffle
from moptipy.operators.permutations.op1_swap2 import Op1Swap2
from moptipy.spaces.permutations import Permutations
from moptipy.utils.console import logger
from moptipy.utils.nputils import (
    DEFAULT_FLOAT,
    DEFAULT_INT,
    rand_generator,
    rand_seeds_from_str,
)
from moptipy.utils.types import check_int_range, type_error


def __can_solve_instance(inst: Instance, seed: int,
                         queue: mp.Queue) -> None:
    """
    Execute one run of a RLS for 2.1min on the instance, return makespan.

    :param inst: the jssp instance to check
    :param seed: the seed of this run
    :param queue: the queue for the return values
    """
    x_space: Final[Permutations] = Permutations.with_repetitions(
        inst.jobs, inst.machines)
    y_space: Final[GanttSpace] = GanttSpace(inst)
    g: Final[OperationBasedEncoding] = OperationBasedEncoding(inst)
    o0: Final[Op0Shuffle] = Op0Shuffle(x_space)
    o1: Final[Op1Swap2] = Op1Swap2()
    f: Final[Makespan] = Makespan(inst)
    a: Final[RLS] = RLS(o0, o1)
    goal: Final[int] = inst.makespan_lower_bound

    ex: Final[Execution] = Execution()
    ex.set_search_space(x_space)
    ex.set_solution_space(y_space)
    ex.set_encoding(g)
    ex.set_algorithm(a)
    ex.set_objective(f)
    ex.set_max_time_millis(126_000)
    ex.set_goal_f(goal)
    ex.set_rand_seed(seed)
    with ex.execute() as process:
        queue.put(process.get_best_f())


def __is_instance_too_easy(inst: Instance) -> bool:
    """
    Check if an instance is too easy to be of any interest.

    We perform 16 runs of a (1+1)-EA on an instance, granting 2.1 minutes
    of runtime to each of them. An instance is considered as easy if one of
    the following conditions is fulfilled:

    - it is solved often enough (lower bound reached often)
    - there are not enough different results
    - the mean result is too close to the lower bound
    - the worst result is not far enough from the lower bound
    - the worst result is too close to the best result

    If any of these conditions hold, we consider the instance as too easy to
    be of any interest for our educational experiment.

    :param inst: the jssp instance to check
    :returns: `True` if the instance is too easy to be interesting
    """
    n_runs: Final[int] = 16
    seeds: Final[list[int]] = rand_seeds_from_str(string=inst.name,
                                                  n_seeds=n_runs)
    queue: Final[mp.Queue] = mp.Queue()

    while len(seeds) > 0:
        processes = [mp.Process(target=__can_solve_instance,
                                args=(inst, seeds.pop(), queue))
                     for _ in range(min(len(seeds), __DEFAULT_N_THREADS))]
        for p in processes:
            p.start()
        for p in processes:
            p.join()

    goal: Final[int] = inst.makespan_lower_bound
    num_solved: int = 0
    diff: set[int] = set()
    min_makespan: int = 1 << 62
    sum_makespans: int = 0
    max_makespan: int = 0

    for _ in range(n_runs):
        makespan: int = queue.get()
        diff.add(makespan)
        if makespan <= min_makespan:
            min_makespan = makespan
            if makespan <= goal:
                num_solved += 1
        sum_makespans += makespan
        if makespan > max_makespan:
            max_makespan = makespan

    mean_makespan: Final[float] = sum_makespans / n_runs
    mean_threshold: Final[float] = min(goal + 32.0,
                                       max(1.01 * goal, goal + 3))
    max_threshold: Final[float] = min(goal + 64.0,
                                      max(1.02 * goal, goal + 5))
    min_threshold: Final[float] = min(min_makespan + 48.0,
                                      max(1.015 * min_makespan,
                                          min_makespan + 4))
    n_diff: Final[int] = len(diff)
    diff_threshold: Final[int] = max(3, int(0.5 + ceil(log2(n_runs))))
    solved_threshold: Final[int] = int(ceil(n_runs / 3))

    result: Final[bool] = (num_solved >= solved_threshold) \
        or (mean_makespan < mean_threshold) \
        or (max_makespan < max_threshold) \
        or (min_threshold > max_makespan) \
        or (n_diff <= diff_threshold)
    logger(f"Instance {inst.name} is {'easy' if result else 'hard'}, "
           f"since {n_runs} runs of the (1+1)-EA had {n_diff} different "
           f"results (threshold {diff_threshold}), a best result of "
           f"{min_makespan}, a mean result of {mean_makespan} (threshold: "
           f"{mean_threshold}), a worst result of {max_makespan} "
           f"(thresholds: {max_threshold}, {min_threshold}), and "
           f"reached the makespan lower bound ({goal}) {num_solved} "
           f"times (threshold: {solved_threshold}).")
    return result


#: The default number of threads to be used
__DEFAULT_N_THREADS: Final[int] = max(1, min(len(sched_getaffinity(0)), 128))


def compute_instances_that_are_too_easy() -> tuple[str, ...]:
    """
    Compute the set of instances that are too easy to be of any interest.

    For our educational experiments, we need instances that are not always
    solved to optimality by simple algorithms. Otherwise, for example, the
    box plots of the end results will collapse to single lines. Then, it
    will also be impossible to really say which algorithm performs best on
    the instance.
    See the documentation of `__is_instance_too_easy` for the conditions that
    lead to the rejection of an instance

    :returns: the tuple of instance names that should not be included in
        the experiment
    """
    logger("We now test the instances whether they are too easy.")
    easy: Final[list[str]] = []
    for inst_name in Instance.list_resources():
        if __is_instance_too_easy(Instance.from_resource(inst_name)):
            easy.append(inst_name)
    easy.sort()
    logger(f"Finished the test, found {len(easy)} easy instances: {easy}.")
    return tuple(easy)


#: This is the set of instances that are too easy to be interesting and shall
#: therefore be skipped from our experiments.
#: For a documentation about what constitutes 'too easy', see
#: :func:`compute_instances_that_are_too_easy`
__TOO_EASY: Final[set[str]] = \
    {"demo", "dmu21", "dmu22", "dmu24", "dmu25", "dmu31", "dmu32", "dmu33",
     "dmu34", "dmu35", "ft06", "ft20", "la01", "la02", "la03", "la04",
     "la05", "la06", "la07", "la08", "la09", "la10", "la11", "la12", "la13",
     "la14", "la15", "la17", "la18", "la23", "la26", "la28", "la30", "la31",
     "la32", "la33", "la34", "la35", "swv16", "swv17", "swv18", "swv19",
     "swv20", "ta51", "ta52", "ta53", "ta54", "ta55", "ta56", "ta57", "ta58",
     "ta59", "ta60", "ta61", "ta63", "ta68", "ta69", "ta71", "ta72", "ta73",
     "ta74", "ta75", "ta76", "ta77", "ta78", "ta79", "ta80"}


def __get_instances() -> list[Instance]:
    """
    Get the instances.

    :return: a tuple of instances
    """
    return [Instance.from_resource(name) for name in
            Instance.list_resources() if name not in __TOO_EASY]


def __optimize_clusters(cluster_groups: tuple[tuple[int, ...], ...],
                        n_groups: int,
                        extreme_groups: tuple[tuple[tuple[int,
                                                          int], ...], ...],
                        random: Generator) -> tuple[int, ...]:
    """
    Try to find optimal cluster-to-group assignments.

    We perform hill climbing with restarts.

    :param cluster_groups: the groups per cluster
    :param n_groups: the number of groups
    :param extreme_groups: the extreme groups
    :param random: the random number generator
    :return: the selected groups
    """
    n: Final[int] = len(cluster_groups)
    current: Final[np.ndarray] = np.zeros(n, DEFAULT_INT)
    current_f: tuple[int, int, int, int, float]
    best: Final[np.ndarray] = np.zeros(n, DEFAULT_INT)
    best_f: tuple[int, int, int, int, float]
    total_best: Final[np.ndarray] = np.zeros(n, DEFAULT_INT)
    total_best_f: tuple[int, int, int, int, float] = (-1, -1, -1, -1, -1)
    run_last_improved: int = 1
    run_current: int = 0
    run_max_none_improved: Final[int] = 4
    step_max_none_improved: Final[int] = int((2 + (n_groups * n)) ** 2)

    done: Final[np.ndarray] = np.zeros(n_groups, DEFAULT_INT)  # noqa
    extremes: Final[set[int]] = set()  # noqa

    logger(f"Beginning to optimize the assignment of {len(cluster_groups)} "
           f"clusters to {n_groups} groups. The minimum groups are "
           f"{extreme_groups[0]}, the maximum groups are {extreme_groups[1]}."
           f" We permit {run_max_none_improved} runs without improvement "
           f"before termination and {step_max_none_improved} FEs without "
           "improvement before stopping a run.")

    def __f(sol: np.ndarray) -> tuple[int, int, int, int, float]:
        """
        Compute the quality: The internal objective function.

        We maximize five things:
        First, we maximize the number of group assignments that allow us to
        include the extreme instances in the result. These are the smallest
        and largest-scale instances.
        Second, we maximize the number of instance sets covered.
        Third, we maximize the minimum number of usages of the instance sets.
        Fourth, we minimize the maximum number of usages of the instance sets
        with the goal to balance the usages evenly over different sets.
        Fifth, we minimize the standard deviation of the instance set usages,
        again with the goal to balance the usages evenly over different sets.

        :param sol: the solution
        :return: the objective values
        """
        nonlocal done  # the number of uses per set
        nonlocal extreme_groups  # the tuple-tuple with the extreme groups
        nonlocal extremes  # the extreme groups already picked

        # first, we count how often each group has been used
        done.fill(0)
        for group in sol:
            done[group] += 1

        # second, we check which cluster-to-group assignment permits
        # using an extreme instance
        extremes.clear()
        for groups in extreme_groups:
            for group in groups:
                if (sol[group[0]] == group[1]) and (group[0] not in extremes):
                    extremes.add(group[0])
                    break
        return len(extremes), int(np.sum(done > 0)), done.min(), \
            -done.max(), -np.std(done)

    # The outer loop: the restarts of the hill climber.
    # We continue to execute runs until a given number of successive runs did
    # not improve the overall best solutions. Then we stop.
    while (run_current - run_last_improved) <= run_max_none_improved:
        run_current += 1

        # generate an initial best solution
        for i in range(n):
            cg = cluster_groups[i]
            best[i] = cg[random.integers(len(cg))]
        best_f = __f(best)  # compute quality
        use_f = best_f

        if best_f > total_best_f:  # accept solution if it is better
            total_best_f = best_f
            run_last_improved = run_current  # new global best
            np.copyto(total_best, best)
            logger(f"New global best {tuple(best)} with quality "
                   f"{best_f} found in run {run_current}.")
        else:
            logger(f"Run {run_current} starts with quality {best_f}.")

        # local search step
        # We continue the local search until a given number of successive
        # steps did not yield any improvement. Then we terminate the run.
        step_current: int = 0
        step_last_improved: int = 1
        while (step_current - step_last_improved) <= step_max_none_improved:
            step_current += 1
            np.copyto(current, best)
            while True:  # perform at least one move
                i = random.integers(n)
                cg = cluster_groups[i]
                lcg = len(cg)
                if lcg <= 1:
                    continue
                while True:  # perform the search move
                    new_v = cg[random.integers(lcg)]
                    if new_v != best[i]:
                        current[i] = new_v
                        break
                # 50/50 chance to perform more moves
                if random.integers(2) <= 0:
                    break

            current_f = __f(current)  # evaluate quality
            # accept if better or with small probability
            if (current_f >= use_f) or (random.integers(50) < 1):
                use_f = current_f
                np.copyto(best, current)
                if current_f > best_f:
                    best_f = current_f
                    step_last_improved = step_current
                    if current_f > total_best_f:  # new optimum
                        total_best_f = current_f
                        run_last_improved = run_current
                        np.copyto(total_best, current)
                        logger(f"New global best {tuple(best)} with quality"
                               f" {total_best_f} found in run {run_current} "
                               f"after {step_current} FEs.")
        logger(f"Run {run_current} finished with quality {best_f}.")

    result = tuple(total_best)
    logger(f"Finished after {run_current} runs with solution {result} of "
           f"quality {total_best_f}.")
    return result


def __optimize_scales(scale_choices: list[list[tuple[int, int]]],
                      random: Generator) -> list[int]:
    """
    Pick a diverse scale choice.

    :param scale_choices: the scale choices we have per group
    :param random: the random number generator
    :returns: one chosen scale per group
    """
    n: Final[int] = len(scale_choices)

    x_total_best: list[int] = [0] * n
    opt_idx: list[tuple[int, int]] = []  # the relevant indexes
    opt_sum: int = 0
    for idx, choices in enumerate(scale_choices):
        if len(choices) > 1:
            opt_sum += len(choices)
            opt_idx.append((idx, len(choices)))

    if not opt_idx:
        return x_total_best

    def __f(xx: list[int]) -> tuple[float, float, int]:
        """
        Compute the quality of a suggestion, bigger is better.

        :param xx: the candidate solution
        :returns: a tuple with minimum distance, distance sum, and
            difference count
        """
        dist_min: float = inf
        dist_sum: float = 0
        diff_cnt: int = 0
        nonlocal scale_choices
        nonlocal n
        for i in range(n):
            a = scale_choices[i][xx[i]]
            for j in range(i):
                b = scale_choices[j][xx[j]]
                if a[0] == b[0]:
                    d0 = 0
                else:
                    diff_cnt += 1
                    d0 = a[0] - b[0]
                if a[1] == b[1]:
                    d1 = 0
                else:
                    diff_cnt += 1
                    d1 = a[1] - b[1]
                d = sqrt((d0 * d0) + (d1 * d1))
                if d < dist_min:
                    dist_min = d
                dist_sum += d
        return dist_min, dist_sum, diff_cnt

    f_total_best: tuple[float, float, int] = __f(x_total_best)
    logger(f"The following index-choices tuple exist: {opt_idx} and "
           f"the initial choice is {x_total_best} at quality {f_total_best}.")

    x_cur_best: list[int] = [0] * n
    x_cur: list[int] = [0] * n

    for _ in range(int(opt_sum ** 2.35)):
        for ii, sc in enumerate(scale_choices):
            x_cur_best[ii] = random.integers(len(sc))
        f_cur_best: tuple[float, float, int] = __f(x_cur_best)
        if f_cur_best > f_total_best:
            f_total_best = f_cur_best
            x_total_best.clear()
            x_total_best.extend(x_cur_best)
            logger(f"Improved to {x_total_best} at quality {f_total_best}.")
        for _ in range(int(opt_sum ** 2.35)):
            idx, choicen = opt_idx[random.integers(len(opt_idx))]
            old = x_cur[idx]
            while old == x_cur[idx]:
                x_cur[idx] = random.integers(choicen)
            f_cur = __f(x_cur)
            if f_cur >= f_cur_best:
                f_cur_best = f_cur
                x_cur_best.clear()
                x_cur_best.extend(x_cur)
                if f_cur_best > f_total_best:
                    f_total_best = f_cur_best
                    x_total_best.clear()
                    x_total_best.extend(x_cur_best)
                    logger(f"Improved to {x_total_best} "
                           f"at quality {f_total_best}.")

    return x_total_best


def __scale(jobs: int, machines: int) -> int:
    """
    Compute the scale of a JSSP instance.

    :param jobs: the jobs
    :param machines: the machines
    :returns: the scale
    """
    return factorial(jobs) ** machines


def propose_instances(n: int,
                      get_instances: Callable = __get_instances) -> \
        tuple[str, ...]:
    """
    Propose a set of instances to be used for our experiment.

    This function is used to obtain the instances chosen for the JSSP
    experiment. You can also use it for other experiments, by using your own
    instance source and/or for selecting more or less JSSP instances.

    Basically, it will accept a function `get_instances`, which must produce
    a sequence of Instance objects. Each instance has a name (e.g., `dmu14`)
    and a group, where the group is the name without any numerical suffix
    (e.g., `dmu`). This function will then select `n` instances from the
    instance set with the goal to maximize the diversity of the instances, to
    include instances from as many groups as possible, and to include one
    instance of the smallest and one of the largest scale.
    The diversity is measured in terms of the numbers of jobs and machines,
    the instance scale, the minimum and maximum operation length, the
    standard deviation of the mean operation lengths over the jobs, the
    makespan bounds, and so on.

    First, features are computed for each instance. Second, the instances are
    clustered into `n` clusters. Third, we try to pick groups for each cluster
    such that a) the minimum and maximum-scale instances can be included and
    b) that instances from as many groups as possible are picked. Third, we
    then randomly pick one instance for each cluster from the selected group
    (while trying to pick the minimum and maximum-scale instances). Finally,
    the chosen instance names are listed as sorted tuple and returned.

    :param n: the number of instances to be proposed
    :param get_instances: a function returning an
        iterable of instances.
    :return: a tuple with the instance names
    """
    check_int_range(n, "n", 1, 1000)
    if not callable(get_instances):
        raise type_error(get_instances, "get_instances", call=True)
    instances = list(get_instances())
    n_instances: Final[int] = len(instances)
    if n_instances <= 0:
        raise ValueError("No instance returned by get_instance.")
    for i in instances:
        if isinstance(i, Instance):
            raise type_error(
                i, "value in list returned by get_instances", Instance)
    if n_instances <= n:
        if n_instances == n:  # nothing to do here
            return tuple(sorted([inst.get_name() for inst in instances]))
        raise ValueError(f"{n} instances requested, but only "
                         f"{n_instances} available.")
    logger(f"We will pick {n} instances out of {n_instances}.")

    random: Final[Generator] = rand_generator(0)  # the random number generator
    random.shuffle(instances)

    inst_names: Final[tuple[str, ...]] = tuple([  # get the instance names
        inst.get_name() for inst in instances])
    inst_sizes: Final[list[tuple[int, int]]] =\
        [(inst.jobs, inst.machines) for inst in instances]

    # the group to which the instances belong
    rm = str.maketrans("", "", "0123456789")
    inst_groups: Final[tuple[str, ...]] = tuple([  # get the instance groups
        inst.translate(rm) for inst in inst_names])

    # create bi-directional mapping between group names and integer IDs
    group_ids: Final[dict[str, int]] = {}
    id_groups: Final[dict[int, str]] = {}
    for group in inst_groups:
        if group in group_ids:
            continue
        gid = len(group_ids)
        group_ids[group] = gid
        id_groups[gid] = group
    n_groups: Final[int] = len(group_ids)
    logger(f"The {n_instances} instances belong to the {n_groups} "
           f"groups {sorted(group_ids.keys())}.")

    # compute the feature matrix
    base_features: Final[int] = 9
    features = np.zeros((len(instances), base_features + n_groups),
                        DEFAULT_FLOAT)
    min_scale_val = inf
    min_scale_inst: set[int] = set()
    max_scale_val = -inf
    max_scale_inst: set[int] = set()
    for i, inst in enumerate(instances):
        features[i, 0] = inst.jobs
        features[i, 1] = inst.machines
        features[i, 2] = inst.jobs / inst.machines
        scale = __scale(inst.jobs, inst.machines)
        if scale <= min_scale_val:
            if scale < min_scale_val:
                min_scale_val = scale
                min_scale_inst.clear()
            min_scale_inst.add(i)
        if scale >= max_scale_val:
            if scale > max_scale_val:
                max_scale_inst.clear()
                max_scale_val = scale
            max_scale_inst.add(i)
        features[i, 3] = log2(log2(scale))
        m = inst.matrix[:, 1::2]
        features[i, 4] = np.std(np.mean(m, axis=0))
        features[i, 5] = np.min(m)
        features[i, 6] = np.max(m)
        features[i, 7] = inst.makespan_lower_bound
        features[i, 8] = inst.makespan_upper_bound / inst.makespan_lower_bound
        features[i, base_features + group_ids[inst_groups[i]]] = 1
    del instances
    logger(f"We computed {base_features + n_groups} features for each "
           f"instance, namely {base_features} features and {n_groups}"
           f" group features. The instances with the smallest scale "
           f" {min_scale_val} are {[inst_names[i] for i in min_scale_inst]} "
           f" (encoded as {min_scale_inst}) and those of the largest scale "
           f"{max_scale_val} are {[inst_names[i] for i in max_scale_inst]} "
           f"(encoded as {max_scale_inst}). Now we will cluster the "
           f"instances.")

    # standardize feature columns
    for i in range(base_features):
        features[:, i] = (features[:, i] - np.mean(features[:, i])) \
            / np.std(features[:, i])

    # so now we have the following things:
    # 1. a list `instances` of instance names
    # 2. the list `instance_groups` with the corresponding groups
    # 3. set `groups` with the group names
    # 4. the matrix `features` with instance features
    # 5. the bi-directional mapping between instance groups and group IDs
    # 6. the instance/group indexes for the smallest and largest-scale
    #    instances

    # now we cluster the instances
    model = SpectralClustering(n_clusters=n,
                               n_init=100,
                               random_state=RandomState(
                                   random.integers(2 ** 32)))
    clusters = model.fit_predict(features)
    if len(clusters) != n_instances:
        raise ValueError(
            f"Invalid number {len(clusters)} of cluster "
            f"assignments to {n_instances} instances.")
    if (max(clusters) - min(clusters) + 1) != n:
        raise ValueError(f"Expected {n} clusters, but got "
                         f"{max(clusters) - min(clusters) + 1}.")
    logger(f"Found clusters {clusters}. The minimum instances are in"
           f"{[clusters[i] for i in min_scale_inst]}. The maximum instances "
           f"are in {[clusters[i] for i in max_scale_inst]}. now assigning"
           f"clusters to groups.")

    # find which groups belong to which cluster
    cluster_groups: Final[tuple[tuple[int, ...], ...]] = tuple([
        tuple(sorted({group_ids[inst_groups[j]]
                      for j in np.where(clusters == i)[0]}))
        for i in range(n)])
    logger(f"The groups available per cluster are {cluster_groups}.")

    # Now we need to pick the extreme groups.
    extreme_groups = tuple(tuple(sorted(
        {(clusters[xx], group_ids[inst_groups[xx]]) for xx in ex}))
        for ex in [min_scale_inst, max_scale_inst])
    logger(f"The extreme groups are {extreme_groups}.")

    # With this method, we choose one instance group for each cluster.
    chosen_groups = __optimize_clusters(cluster_groups=cluster_groups,
                                        extreme_groups=extreme_groups,
                                        n_groups=n_groups,
                                        random=random)
    logger(f"The instance groups {chosen_groups} were chosen for the"
           f" {n} clusters.")

    # OK, we have picked one instance group per cluster.
    # Now we need to pick the instance scales from these groups.
    # The goal is to now select instances with (jobs, machines) settings as
    # diverse as possible, while adhering to the selected instance groups.
    # For example, a selected group may be of the family "la", but there
    # might be instances of different (job, machines) settings in the cluster
    # for this group.
    # In this case, we want to pick those which do not already occur in other
    # clusters.
    # If we can, we will pick the instances with the minimum and the maximum
    # scales.
    # In a first step, we find out
    needs_min_scale = True
    needs_max_scale = True
    scale_choices: list[list[tuple[int, int]]] = []
    inst_choices: list[list[str]] = []

    for i in range(n):
        elements = np.where(clusters == i)[0]
        sgroup: str = id_groups[chosen_groups[i]]
        possibility: set[int] = {i for i in elements
                                 if inst_groups[i] == sgroup}
        cur_scale_choices: list[tuple[int, int]] = []
        scale_choices.append(cur_scale_choices)
        cur_inst_choices: list[str] = []
        inst_choices.append(cur_inst_choices)
        can_skip: bool = False

        if needs_min_scale:
            test = possibility.intersection(min_scale_inst)
            if len(test) > 0:
                logger(
                    f"Choosing from groups {[inst_names[t] for t in test]} "
                    f"for cluster {i} to facilitate minimum-scale instance.")
                sel = list(test)
                seli = sel[random.integers(len(sel))]
                possibility = {seli}
                cur_scale_choices.append(inst_sizes[seli])
                cur_inst_choices.append(inst_names[seli])
                can_skip = True
                needs_min_scale = False

        if needs_max_scale:
            test = possibility.intersection(max_scale_inst)
            if len(test) > 0:
                logger(
                    f"Choosing from groups {[inst_names[t] for t in test]} "
                    f"for cluster {i} to facilitate maximum-scale instance.")
                needs_max_scale = False
                if can_skip:
                    continue
                sel = list(test)
                seli = sel[random.integers(len(sel))]
                cur_scale_choices.append(inst_sizes[seli])
                cur_inst_choices.append(inst_names[seli])
                continue
        if can_skip:
            continue

        scales: set[tuple[int, int]] = set()
        sel = list(possibility)
        random.shuffle(sel)
        for ii in sel:
            tup = inst_sizes[ii]
            if tup not in scales:
                scales.add(tup)
                cur_inst_choices.append(inst_names[ii])
                cur_scale_choices.append(tup)

    logger(f"Got the scale choices {scale_choices} resulting from the "
           f"possible instances {inst_choices}.")
    # do the actual scale optimization
    final_sel = __optimize_scales(scale_choices, random)

    res_tmp: Final[list[tuple[int, str]]] = \
        [(__scale(scale_choices[i][k][0], scale_choices[i][k][1]),
          inst_choices[i][k]) for i, k in enumerate(final_sel)]
    res_tmp.sort()
    result: Final[list[str]] = [sey[1] for sey in res_tmp]

    # Finally, we sort and finalize the set of chosen instances.
    logger(f"Found final instance selection {result}.")
    return tuple(result)



"""Some basic type handling routines."""
from typing import Any, Final, Iterable


def type_name(tpe: type) -> str:
    """
    Convert a type to a string.

    :param tpe: the type
    :returns: the string

    >>> type_name(type(None))
    'None'
    >>> type_name(int)
    'int'
    >>> from moptipy.utils.path import Path
    >>> type_name(Path)
    'moptipy.utils.path.Path'
    >>> from typing import Callable
    >>> type_name(Callable)
    'typing.Callable'
    >>> from typing import Callable as Ca
    >>> type_name(Ca)
    'typing.Callable'
    >>> from typing import Callable as Aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
    >>> type_name(Aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa)
    'typing.Callable'
    >>> import typing as ttttttttttttttttttttttttttttttttttttttttttt
    >>> type_name(ttttttttttttttttttttttttttttttttttttttttttt.Callable)
    'typing.Callable'
    >>> import numpy.random
    >>> type_name(numpy.random.Generator)
    'numpy.random._generator.Generator'
    >>> import numpy.random as npr
    >>> type_name(npr.Generator)
    'numpy.random._generator.Generator'
    >>> from numpy.random import Generator as Ge
    >>> type_name(Ge)
    'numpy.random._generator.Generator'
    >>> import numpy as npx
    >>> type_name(npx.random.Generator)
    'numpy.random._generator.Generator'
    """
    c1: str = str(tpe)
    if c1.startswith("<class '"):
        c1 = c1[8:-2]
    if c1 == "NoneType":
        return "None"
    return c1


def type_name_of(obj) -> str:
    """
    Get the fully-qualified class name of an object.

    :param obj: the object
    :returns: the fully-qualified class name of the object

    >>> from moptipy.examples.jssp.instance import Instance
    >>> type_name_of(Instance.from_resource("demo"))
    'moptipy.examples.jssp.instance.Instance'
    >>> from numpy.random import default_rng
    >>> type_name_of(default_rng())
    'numpy.random._generator.Generator'
    >>> type_name_of(None)
    'None'
    >>> type_name_of(int)
    'type'
    >>> type_name_of(print)
    'builtin_function_or_method'
    >>> import numpy as npx
    >>> type_name_of(npx)
    'module'
    >>> type_name_of(npx.ndarray)
    'numpy.type'
    >>> from typing import Callable
    >>> type_name_of(Callable)
    'typing._CallableType'
    >>> from math import sin
    >>> type_name_of(sin)
    'math.builtin_function_or_method'
    """
    if obj is None:
        return "None"
    c1: Final[str] = type_name(type(obj))
    if hasattr(obj, "__class__"):
        cls: Final[type] = obj.__class__
        c2: str = type_name(cls)
        if hasattr(cls, "__qualname__"):
            c3: str = cls.__qualname__
            if hasattr(obj, "__module__"):
                module = obj.__module__
                if (module is not None) and (module != "builtins"):
                    c3 = f"{module}.{c3}"
            if len(c3) > len(c2):
                c2 = c3
        if len(c2) > len(c1):
            return c2
    return c1


def type_error(obj: Any, name: str,
               expected: None | type | Iterable[type] = None,
               call: bool = False) -> ValueError | TypeError:
    """
    Create an error to raise if a type did not fit.

    :param obj: the object that is of the wrong type
    :param name: the name of the object
    :param expected: the expected types (or `None`)
    :param call: the object should have been callable?
    :returns: a :class:`TypeError` with a descriptive information

    >>> type_error(1.3, "var", int)
    TypeError("var should be an instance of int but is float, namely '1.3'.")
    >>> type_error("x", "z", (int, float)).args[0]
    "z should be an instance of any in {float, int} but is str, namely 'x'."
    >>> type_error("f", "q", call=True).args[0]
    "q should be a callable but is str, namely 'f'."
    >>> type_error("1", "2", bool, call=True).args[0]
    "2 should be an instance of bool or a callable but is str, namely '1'."
    >>> type_error(None, "x", str)
    TypeError('x should be an instance of str but is None.')
    """
    exp: str = ""
    if isinstance(expected, Iterable):
        exp = ", ".join(sorted([type_name(e) for e in expected]))
        exp = f"an instance of any in {{{exp}}}"
    elif expected is not None:
        exp = f"an instance of {type_name(expected)}"
    if call:
        exp = f"{exp} or a callable" if exp else "a callable"

    message: Final[str] = f"{name} should be {exp} but is None." \
        if obj is None else \
        f"{name} should be {exp} but is {type_name_of(obj)}, " \
        f"namely {str(obj)!r}."

    return TypeError(message)


def check_int_range(val: Any, name: str | None = None,
                    min_value: int | float = 0,
                    max_value: int | float = 1_000_000_000) -> int:
    """
    Check whether a value `val` is an integer in a given range.

    Via type annotation, this method actually accepts a value `val` of any
    type as input. However, if `val` is not an instance of `int`, it will
    throw an error. Also, if `val` is not in the prescribed range, it will
    throw an error, too. By default, the range is `0...1_000_000_000`.

    I noticed that often, we think that  only want to check a lower limit
    for `val`, e.g., that a number of threads or a population size should be
    `val > 0`. However, in such cases, there also always a reasonable upper
    limits. We never actually want an EA to have a population larger than,
    say, 1_000_000_000. That would make no sense. So indeed, whenever we have
    a lower limit for a parameter, we also should have an upper limit
    resulting from physical constraints. 1_000_000_000 is a reasonably sane
    upper limit in many situations. If we need smaller or larger limits, we
    can of course specify them.

    Notice that there is one strange border case: In Python, `bool` is a
    subtype of `int`, where `True` has value `1` and `False` has value `0`.
    See <https://docs.python.org/3/library/functions.html#bool>.
    We therefore treat `bool` values indeed as instances of `int`.

    :param val: the value to check
    :param name: the name of the value, or `None`
    :param min_value: the minimum permitted value
    :param max_value: the maximum permitted value
    :returns: `val` if everything is OK
    :raises TypeError: if `val` is not an `int`
    :raises ValueError: if `val` is an `int` but outside the prescribed range

    >>> try:
    ...   print(check_int_range(12, min_value=7, max_value=13))
    ... except (ValueError, TypeError) as err:
    ...   print(err)
    12

    >>> try:
    ...   print(check_int_range(123, min_value=7, max_value=13))
    ... except (ValueError, TypeError) as err:
    ...   print(err)
    ...   print(err.__class__)
    Value=123 is invalid, must be in 7..13.
    <class 'ValueError'>

    >>> try:
    ...   print(check_int_range(5.0, name="ThisIsFloat"))
    ... except (ValueError, TypeError) as err:
    ...   print(err)
    ...   print(err.__class__)
    ThisIsFloat should be an instance of int but is float, namely '5.0'.
    <class 'TypeError'>

    The behavior in the border case of `bool` instances actually also being
    instances of `int`:

    >>> check_int_range(True, "true", 0, 2)
    True

    >>> check_int_range(False, "false", 0, 2)
    False

    >>> try:
    ...   print(check_int_range(True, min_value=7, max_value=13))
    ... except (ValueError, TypeError) as err:
    ...   print(err)
    ...   print(err.__class__)
    Value=True is invalid, must be in 7..13.
    <class 'ValueError'>
    """
    if not isinstance(val, int):
        raise type_error(val, "value" if name is None else name, int)
    if min_value <= val <= max_value:
        return val
    raise ValueError(f"{'Value' if name is None else name}={val!r} is "
                     f"invalid, must be in {min_value}..{max_value}.")


def check_to_int_range(val: Any, name: str | None = None,
                       min_value: int | float = 0,
                       max_value: int | float = 1_000_000_000) -> int:
    """
    Check whether a value `val` can be converted an integer in a given range.

    :param val: the value to convert via `int(...)` and then to check
    :param name: the name of the value, or `None`
    :param min_value: the minimum permitted value
    :param max_value: the maximum permitted value
    :returns: `val` if everything is OK
    :raises TypeError: if `val` is `None`
    :raises ValueError: if `val` is not `None` but can either not be converted
       to an `int` or to an `int` outside the prescribed range

    >>> try:
    ...   print(check_to_int_range(12))
    ... except (ValueError, TypeError) as err:
    ...   print(err)
    12

    >>> try:
    ...   print(check_to_int_range(12.0))
    ... except (ValueError, TypeError) as err:
    ...   print(err)
    12

    >>> try:
    ...   print(check_to_int_range("12"))
    ... except (ValueError, TypeError) as err:
    ...   print(err)
    12

    >>> try:
    ...   print(check_to_int_range("A"))
    ... except (ValueError, TypeError) as err:
    ...   print(err)
    ...   print(err.__class__)
    Cannot convert value='A' to int, let alone in range 0..1000000000.
    <class 'ValueError'>

    >>> try:
    ...   print(check_to_int_range(None))
    ... except (ValueError, TypeError) as err:
    ...   print(err)
    ...   print(err.__class__)
    Cannot convert value=None to int, let alone in range 0..1000000000.
    <class 'TypeError'>
    """
    try:
        conv = int(val)
    except (ValueError, TypeError) as errx:
        raise (ValueError if isinstance(errx, ValueError) else TypeError)(
            f"Cannot convert {'value' if name is None else name}={val!r} "
            f"to int, let alone in range {min_value}..{max_value}.") from errx
    return check_int_range(conv, name, min_value, max_value)



"""
A fitness representing the rank of a solution based on its quality.

First, all solutions are sorted based on their objective values. Then, they
receive their rank, i.e., index+1 in the sorted list, as fitness. If two
elements have the same objective value, they also receive the same rank.
This rank will be the index+1 of the first element.

>>> # The first value in FRecord(x, y) is the solution, which is not matter
>>> # here. The second value is the objective value used for ranking.
>>> l = [FRecord(1, 10), FRecord(2, 3), FRecord(3, 3), FRecord(4, 2)]
>>> from numpy.random import default_rng
>>> Rank().assign_fitness(l, default_rng())
>>> for z in l:
...     print(f"x={z.x}, fitness={z.fitness}")
x=4, fitness=1
x=2, fitness=2
x=3, fitness=2
x=1, fitness=4

Together with fitness proportionate selection (:class:`~moptipy.algorithms.\
modules.selections.fitness_proportionate_sus.FitnessProportionateSUS`), it
works very similar to linear ranking selection in an EA
(:class:`~moptipy.algorithms.so.general_ea.GeneralEA`).

1. L. Darrell Whitley. The GENITOR Algorithm and Selection Pressure: Why
   Rank-Based Allocation of Reproductive Trials is Best. In J. David Schaffer,
   ed., Proceedings of the 3rd International Conference on Genetic Algorithms
   (ICGA'89), June 4-7, 1989, Fairfax, VA, USA, pages 116-121. San Francisco,
   CA, USA: Morgan Kaufmann Publishers Inc. ISBN: 1-55860-066-3
   https://www.researchgate.net/publication/2527551
2. Tobias Blickle and Lothar Thiele. A Comparison of Selection Schemes used in
   Genetic Algorithms. Second edition, December 1995. TIK-Report 11 from the
   Eidgenössische Technische Hochschule (ETH) Zürich, Department of Electrical
   Engineering, Computer Engineering and Networks Laboratory (TIK), Zürich,
   Switzerland. ftp://ftp.tik.ee.ethz.ch/pub/publications/TIK-Report11.ps
"""

from math import inf

from numpy.random import Generator

from moptipy.algorithms.so.fitness import Fitness, FRecord


# start book
class Rank(Fitness):
    """A fitness computing the rank of an individual based on its quality."""

    def assign_fitness(self, p: list[FRecord], random: Generator) -> None:
        """
        Assign the rank as fitness.

        :param p: the list of records
        :param random: ignored

        >>> l = [FRecord(0, 10), FRecord(1, 5), FRecord(2, 5), FRecord(3, -1)]
        >>> from numpy.random import default_rng
        >>> Rank().assign_fitness(l, default_rng())
        >>> l[0].x
        3
        >>> l[0].fitness
        1
        >>> l[1].x
        1
        >>> l[1].fitness
        2
        >>> l[2].x
        2
        >>> l[2].fitness
        2
        >>> l[3].x
        0
        >>> l[3].fitness
        4
        """
        for rec in p:  # first copy rec.f to rec.fitness
            rec.fitness = rec.f  # because then we can easily sort
        p.sort()  # sort based on objective values

        rank: int = -1  # the variable for storing the current rank
        last_f: int | float = -inf  # the previous objective value
        for i, rec in enumerate(p):  # iterate over list
            v = rec.fitness  # get the current objective value
            if v > last_f:  # only increase rank if objective f changes
                rank = i + 1  # +1 so smallest-possible fitness is 1
                last_f = v  # remember objective value for comparison
            rec.fitness = rank  # assign the rank (same f = same rank)
# end book

    def __str__(self):
        """
        Get the name of this rank-based fitness assignment.

        :return: the name of this fitness assignment strategy
        :retval "rank": always
        """
        return "rank"



"""
Approximate the expected running time to reach certain goals.

The (empirically estimated) Expected Running Time (ERT) tries to give an
impression of how long an algorithm needs to reach a certain solution quality.

The ERT for a problem instance is estimated as the ratio of the sum of all FEs
that all the runs consumed until they either have discovered a solution of a
given goal quality or exhausted their budget, divided by the number of runs
that discovered a solution of the goal quality. The ERT is the mean expect
runtime under the assumption of independent restarts after failed runs, which
then may either succeed (consuming the mean runtime of the successful runs) or
fail again (with the observed failure probability, after consuming the
available budget).

The ERT itself can be considered as a function that associates the estimated
runtime given above to all possible solution qualities that can be attained by
an algorithm for a give problem. For qualities/goals that an algorithm did not
attain in any run, the ERT becomes infinite.

1. Kenneth V. Price. Differential Evolution vs. The Functions of the 2nd ICEO.
   In Russ Eberhart, Peter Angeline, Thomas Back, Zbigniew Michalewicz, and
   Xin Yao, editors, *IEEE International Conference on Evolutionary
   Computation,* April 13-16, 1997, Indianapolis, IN, USA, pages 153-157.
   IEEE Computational Intelligence Society. ISBN: 0-7803-3949-5.
   doi: https://doi.org/10.1109/ICEC.1997.592287
2. Nikolaus Hansen, Anne Auger, Steffen Finck, Raymond Ros. *Real-Parameter
   Black-Box Optimization Benchmarking 2010: Experimental Setup.*
   Research Report RR-7215, INRIA. 2010. inria-00462481.
   https://hal.inria.fr/inria-00462481/document/
"""

from dataclasses import dataclass
from math import inf, isfinite
from typing import Any, Callable, Final, Iterable, cast

import numpy as np

import moptipy.api.logging as lg
import moptipy.utils.nputils as npu
from moptipy.evaluation._utils import _get_reach_index
from moptipy.evaluation.base import (
    F_NAME_NORMALIZED,
    F_NAME_RAW,
    F_NAME_SCALED,
    KEY_N,
    MultiRun2DData,
)
from moptipy.evaluation.progress import Progress
from moptipy.utils.console import logger
from moptipy.utils.logger import (
    COMMENT_CHAR,
    CSV_SEPARATOR,
    KEY_VALUE_SEPARATOR,
)
from moptipy.utils.path import Path
from moptipy.utils.strings import num_to_str
from moptipy.utils.types import type_error


def compute_single_ert(source: Iterable[Progress],
                       goal_f: int | float) -> float:
    """
    Compute a single ERT.

    The ERT is the sum of the time that the runs spend with a
    best-so-far quality greater or equal than `goal_f` divided by the
    number of runs that reached `goal_f`. The idea is that the
    unsuccessful runs spent their complete computational budget and
    once they have terminated, we would immediately start a new,
    independent run.

    :param source: the source array
    :param goal_f: the goal objective value
    :return: the ERT
    """
    n_success: int = 0
    time_sum: int = 0
    for progress in source:
        size = progress.f.size
        idx = size - _get_reach_index(progress.f, goal_f)
        if idx < size:
            n_success += 1
        else:
            idx = cast(np.integer, -1)
        time_sum = time_sum + int(progress.time[idx])
    if time_sum <= 0:
        raise ValueError(f"Time sum cannot be {time_sum}.")
    if n_success <= 0:
        return inf
    return time_sum / n_success


@dataclass(frozen=True, init=False, order=True)
class Ert(MultiRun2DData):
    """Estimate the Expected Running Time (ERT)."""

    #: The ert function
    ert: np.ndarray

    def __init__(self,
                 algorithm: str | None,
                 instance: str | None,
                 n: int,
                 time_unit: str,
                 f_name: str,
                 ert: np.ndarray):
        """
        Create the Ert function.

        :param algorithm: the algorithm name, if all runs are with the same
            algorithm
        :param instance: the instance name, if all runs are on the same
            instance
        :param n: the total number of runs
        :param time_unit: the time unit
        :param f_name: the objective dimension name
        :param ert: the ert matrix
        """
        super().__init__(algorithm, instance, n, time_unit, f_name)
        if not isinstance(ert, np.ndarray):
            raise type_error(ert, "ert", np.ndarray)
        ert.flags.writeable = False

        f: Final[np.ndarray] = ert[:, 0]
        if not npu.is_all_finite(f):
            raise ValueError(
                f"Ert x-axis must be all finite, but encountered {f}.")
        ll = f.size
        if (ll > 1) and np.any(f[1:] <= f[:-1]):
            raise ValueError("f data must be strictly increasing,"
                             f"but encountered {f}.")

        t: Final[np.ndarray] = ert[:, 1]
        if np.isfinite(t[0]) or (np.isposinf(t[0])):
            if not npu.is_all_finite(t[1:]):
                raise ValueError(
                    "non-first ert y-axis elements must be all finite.")
            if np.any(t[1:] >= t[:-1]):
                raise ValueError("t data must be strictly decreasing,"
                                 f"but encountered {t}.")
        else:
            raise ValueError(
                f"first ert y-axis element cannot be {t[0]}.")
        object.__setattr__(self, "ert", ert)

    def to_csv(self, file: str,
               put_header: bool = True) -> Path:
        """
        Store a :class:`Ert` record in a CSV file.

        :param file: the file to generate
        :param put_header: should we put a header with meta-data?
        :return: the fully resolved file name
        """
        path: Final[Path] = Path.path(file)
        logger(f"Writing ERT to CSV file {path!r}.")

        with path.open_for_write() as out:
            sep: Final[str] = CSV_SEPARATOR
            if put_header:
                kv: Final[str] = KEY_VALUE_SEPARATOR
                cmt: Final[str] = COMMENT_CHAR
                if self.algorithm is not None:
                    out.write(
                        f"{cmt} {lg.KEY_ALGORITHM}{kv}{self.algorithm}\n")
                if self.instance is not None:
                    out.write(
                        f"{cmt} {lg.KEY_INSTANCE}{kv}{self.instance}\n")
                out.write(
                    f"{cmt} {KEY_N}{kv}{self.n}\n")
            out.write(f"{self.f_name}{sep}ert[{self.time_unit}]\n")
            for v in self.ert:
                out.write(
                    f"{num_to_str(v[0])}{sep}{num_to_str(v[1])}\n")

        logger(f"Done writing ERT to CSV file {path!r}.")

        path.enforce_file()
        return path

    @staticmethod
    def create(source: Iterable[Progress],
               f_lower_bound: int | float | Callable | None = None,
               use_default_lower_bounds: bool = True) -> "Ert":
        """
        Create one single Ert record from an iterable of Progress records.

        :param source: the set of progress instances
        :param f_lower_bound: the lower bound for the objective value, or a
            callable that is applied to a progress object to get the lower
            bound
        :param use_default_lower_bounds: should we use the default lower
            bounds
        :return: the Ert record
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)

        lower_bound: int | float = inf
        if (f_lower_bound is not None) and (not callable(f_lower_bound)):
            if not isfinite(f_lower_bound):
                raise ValueError("f_lower_bound must be finite "
                                 f"but is {f_lower_bound}.")
            lower_bound = f_lower_bound
            f_lower_bound = None

        algorithm: str | None = None
        instance: str | None = None
        time_unit: str | None = None
        f_name: str | None = None
        f_list: list[np.ndarray] = []
        n: int = 0

        prgs: Final[list[Progress]] = cast(list[Progress], source) \
            if isinstance(source, list) else list(source)

        for progress in prgs:
            if not isinstance(progress, Progress):
                raise type_error(progress, "progress input", Progress)
            if n <= 0:
                algorithm = progress.algorithm
                instance = progress.instance
                time_unit = progress.time_unit
                f_name = progress.f_name
            else:
                if algorithm != progress.algorithm:
                    algorithm = None
                if instance != progress.instance:
                    instance = None
                if time_unit != progress.time_unit:
                    raise ValueError(
                        f"Cannot mix time units {time_unit} "
                        f"and {progress.time_unit}.")
                if f_name != progress.f_name:
                    raise ValueError(f"Cannot mix f-names {f_name} "
                                     f"and {progress.f_name}.")
            n += 1
            if use_default_lower_bounds and \
                    (progress.f_standard is not None) and \
                    (progress.f_name == F_NAME_RAW) and \
                    (lower_bound > progress.f_standard):
                lower_bound = progress.f_standard
            if f_lower_bound is not None:
                lb = f_lower_bound(progress)
                if not isinstance(lb, int | float):
                    raise type_error(lb, "computed lower bound", (int, float))
                if not isfinite(lb):
                    raise ValueError(f"Invalid computed lower bound {lb}.")
                if lb < lower_bound:
                    lower_bound = lb
            f_list.append(progress.f)

        if n <= 0:
            raise ValueError("Did not encounter any progress information.")

        if use_default_lower_bounds:
            if f_name == F_NAME_SCALED:
                lower_bound = min(lower_bound, 1)
            elif f_name == F_NAME_NORMALIZED:
                lower_bound = min(lower_bound, 0)

        # get unique x-values and make sure that lower bound is included
        has_lb: Final[bool] = isfinite(lower_bound)
        if has_lb:
            f_list.append(np.array([lower_bound]))
        x = np.concatenate(f_list)
        del f_list
        x = np.unique(x)
        if has_lb:
            x = x[x >= lower_bound]
        base_len: Final[int] = x.size

        # prepare for backward iteration over arrays
        indices = np.array([(ppr.f.size - 1) for ppr in prgs],
                           dtype=npu.DEFAULT_INT)
        y = np.empty(base_len, dtype=npu.DEFAULT_FLOAT)

        for out_index in range(base_len):
            f_lim = x[out_index]
            found = 0
            time_sum: int = 0
            for r_idx, pr in enumerate(prgs):
                idx = indices[r_idx]
                f_vals = pr.f
                # did we fulfill the limit?
                if f_vals[idx] <= f_lim:
                    found += 1
                    # can we move time cursor back?
                    if (idx > 0) and (f_vals[idx - 1] <= f_lim):
                        idx -= 1
                        indices[r_idx] = idx
                else:
                    # condition not fulfilled, need to use maximum time
                    idx = -1
                time_sum = time_sum + int(pr.time[idx])

            # compute ert value: infinite if no run meets condition
            y[out_index] = inf if (found <= 0) else time_sum / found

        # convert the two arrays into one matrix
        ert = np.concatenate((x, y)).reshape((base_len, 2), order="F")

        return Ert(algorithm, instance, n,
                   time_unit, f_name, ert)

    @staticmethod
    def from_progresses(source: Iterable[Progress],
                        consumer: Callable[["Ert"], Any],
                        f_lower_bound: float | None = None,
                        use_default_lower_bounds: bool = True,
                        join_all_algorithms: bool = False,
                        join_all_instances: bool = False) -> None:
        """
        Compute one or multiple ERTs from a stream of end results.

        :param source: the set of progress instances
        :param f_lower_bound: the lower bound for the objective value
        :param use_default_lower_bounds: should we use the default lower
            bounds
        :param consumer: the destination to which the new records will be
            passed, can be the `append` method of a :class:`list`
        :param join_all_algorithms: should the Ert be aggregated over all
            algorithms
        :param join_all_instances: should the Ert be aggregated over all
            algorithms
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        if not isinstance(join_all_algorithms, bool):
            raise type_error(join_all_algorithms, "join_all_algorithms", bool)
        if not isinstance(join_all_instances, bool):
            raise type_error(join_all_instances, "join_all_instances", bool)

        if join_all_algorithms and join_all_instances:
            consumer(Ert.create(source, f_lower_bound,
                                use_default_lower_bounds))
            return

        sorter: dict[str, list[Progress]] = {}
        for er in source:
            if not isinstance(er, Progress):
                raise type_error(er, "progress source", Progress)
            key = er.instance if join_all_algorithms else \
                er.algorithm if join_all_instances else \
                f"{er.algorithm}/{er.instance}"
            if key in sorter:
                lst = sorter[key]
            else:
                lst = []
                sorter[key] = lst
            lst.append(er)

        if len(sorter) <= 0:
            raise ValueError("source must not be empty")

        if len(sorter) > 1:
            keyz = list(sorter.keys())
            keyz.sort()
            for key in keyz:
                consumer(Ert.create(sorter[key], f_lower_bound,
                                    use_default_lower_bounds))
        else:
            consumer(Ert.create(next(iter(sorter.values())), f_lower_bound,
                                use_default_lower_bounds))



"""Provides function :func:`tabulate_end_results` to tabulate end results."""

from math import inf, isfinite, isnan, nan
from typing import Any, Callable, Final, Iterable, cast

from moptipy.api.logging import (
    KEY_ALGORITHM,
    KEY_BEST_F,
    KEY_GOAL_F,
    KEY_INSTANCE,
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
)
from moptipy.evaluation.base import F_NAME_RAW, F_NAME_SCALED
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.end_statistics import KEY_BEST_F_SCALED, EndStatistics
from moptipy.evaluation.statistics import (
    KEY_MAXIMUM,
    KEY_MEAN_ARITH,
    KEY_MEAN_GEOM,
    KEY_MEDIAN,
    KEY_MINIMUM,
    KEY_STDDEV,
)
from moptipy.utils.formatted_string import FormattedStr
from moptipy.utils.lang import Lang
from moptipy.utils.logger import SCOPE_SEPARATOR
from moptipy.utils.markdown import Markdown
from moptipy.utils.number_renderer import (
    DEFAULT_NUMBER_RENDERER,
    NumberRenderer,
)
from moptipy.utils.path import Path
from moptipy.utils.table import Table
from moptipy.utils.text_format import TextFormatDriver
from moptipy.utils.types import type_error

#: the lower bound key
__KEY_LOWER_BOUND: Final[str] = "lower_bound"
#: the lower bound short key
__KEY_LOWER_BOUND_SHORT: Final[str] = "lower_bound_short"


def default_column_namer(col: str) -> str:
    """
    Get the default name for columns.

    :param col: the column identifier
    :returns: the column name
    """
    if not isinstance(col, str):
        raise type_error(col, "column name", str)
    if col == KEY_INSTANCE:
        return "I"
    if col == KEY_ALGORITHM:
        return Lang.translate("setup")
    if col in (__KEY_LOWER_BOUND, __KEY_LOWER_BOUND_SHORT, KEY_GOAL_F):
        return "lb(f)"
    if col in "summary":
        return Lang.translate(col)

    if SCOPE_SEPARATOR not in col:
        raise ValueError(
            f"statistic {col!r} should contain {SCOPE_SEPARATOR!r}.")
    key, stat = col.split(SCOPE_SEPARATOR)
    if (len(key) <= 0) or (len(stat) <= 0):
        raise ValueError(f"invalid statistic {col!r}.")

    if key == KEY_BEST_F_SCALED:
        key = F_NAME_SCALED
    elif key == KEY_BEST_F:
        key = F_NAME_RAW

    # now fix statistics part
    if stat == KEY_MEAN_ARITH:
        stat = "mean"
    elif stat == KEY_STDDEV:
        stat = "sd"
    elif stat == KEY_MEAN_GEOM:
        stat = "gmean"
    elif stat == KEY_MEDIAN:
        stat = "med"
    elif stat == KEY_MINIMUM:
        stat = Lang.translate("best") \
            if key in (F_NAME_RAW, F_NAME_SCALED) else "min"
    elif stat == KEY_MAXIMUM:
        stat = Lang.translate("worst") \
            if key in (F_NAME_RAW, F_NAME_SCALED) else "max"
    else:
        raise ValueError(f"unknown statistic {stat!r} for {col!r}.")

    if key == F_NAME_RAW:
        return stat
    if key == F_NAME_SCALED:
        return f"{stat}1"
    if key == KEY_LAST_IMPROVEMENT_FE:
        key = "fes"
    elif key == KEY_TOTAL_FES:
        key = "FEs"
    elif key == KEY_LAST_IMPROVEMENT_TIME_MILLIS:
        key = "t"
    elif key == KEY_TOTAL_TIME_MILLIS:
        key = "T"
    else:
        raise ValueError(f"unknown key {key!r}.")
    return f"{stat}({key})"


def command_column_namer(
        col: str, put_dollars: bool = True,
        summary_name: Callable[[bool], str] = lambda put_dollars: r"\summary",
        setup_name: Callable[[bool], str] = lambda put_dollars: r"\setup") \
        -> str:
    """
    Get the command-based names for columns, but in command format.

    This function returns LaTeX-style commands for the column headers.

    :param col: the column identifier
    :param put_dollars: surround the command with `$`
    :param summary_name: the name function for the key "summary"
    :param setup_name: the name function for the key `KEY_ALGORITHM`
    :returns: the column name
    """
    if not isinstance(col, str):
        raise type_error(col, "column name", str)
    if not isinstance(put_dollars, bool):
        raise type_error(put_dollars, "put_dollars", bool)
    if not callable(summary_name):
        raise type_error(summary_name, "summary_name", call=True)
    if not callable(setup_name):
        raise type_error(setup_name, "setup_name", call=True)
    if col == KEY_INSTANCE:
        return r"$\instance$"
    if col == KEY_ALGORITHM:
        return setup_name(put_dollars)
    if col in (__KEY_LOWER_BOUND, __KEY_LOWER_BOUND_SHORT, KEY_GOAL_F):
        return r"$\lowerBound(\objf)$" if \
            put_dollars else r"\lowerBound(\objf)"
    if col == "summary":
        return summary_name(put_dollars)

    if SCOPE_SEPARATOR not in col:
        raise ValueError(
            f"statistic {col!r} should contain {SCOPE_SEPARATOR!r}.")
    key, stat = col.split(SCOPE_SEPARATOR)
    if (len(key) <= 0) or (len(stat) <= 0):
        raise ValueError(f"invalid statistic {col!r}.")

    if key == KEY_BEST_F_SCALED:
        key = F_NAME_SCALED
    elif key == KEY_BEST_F:
        key = F_NAME_RAW

    # now fix statistics part
    if stat == KEY_MEAN_ARITH:
        stat = "mean"
    elif stat == KEY_STDDEV:
        stat = "stddev"
    elif stat == KEY_MEAN_GEOM:
        stat = "geomean"
    elif stat == KEY_MEDIAN:
        stat = "median"
    elif stat == KEY_MINIMUM:
        stat = "min"
    elif stat == KEY_MAXIMUM:
        stat = "max"
    else:
        raise ValueError(f"unknown statistic {stat!r} for {col!r}.")

    if key == F_NAME_RAW:
        key = "BestF"
    elif key == F_NAME_SCALED:
        key = "BestFscaled"
    elif key == KEY_LAST_IMPROVEMENT_FE:
        key = "LIFE"
    elif key == KEY_TOTAL_FES:
        key = "TotalFEs"
    elif key == KEY_LAST_IMPROVEMENT_TIME_MILLIS:
        key = "LIMS"
    elif key == KEY_TOTAL_TIME_MILLIS:
        key = "TotalMS"
    else:
        raise ValueError(f"unknown key {key!r}.")
    return f"$\\{stat}{key}$" if put_dollars else f"\\{stat}{key}"


def __finite_max(data: Iterable[int | float | None]) \
        -> int | float:
    """
    Compute the finite maximum of a data column.

    :param data: the data to iterate over
    :returns: the finite maximum, or `nan` if none can be found or if there
        is only a single value
    """
    if not isinstance(data, Iterable):
        raise type_error(data, "data", Iterable)
    maxi: int | float = -inf
    count: int = 0
    for d in data:
        count += 1
        if d is None:
            continue
        if isfinite(d) and (d > maxi):
            maxi = d
    return maxi if (count > 1) and isfinite(maxi) else nan


def __finite_min(data: Iterable[int | float | None]) \
        -> int | float:
    """
    Compute the finite minimum of a data column.

    :param data: the data to iterate over
    :returns: the finite minimum, or `nan` if none can be found
    """
    if not isinstance(data, Iterable):
        raise type_error(data, "data", Iterable)
    mini: int | float = inf
    count: int = 0
    for d in data:
        count += 1
        if d is None:
            continue
        if isfinite(d) and (d < mini):
            mini = d
    return mini if (count > 1) and isfinite(mini) else nan


def __nan(data: Iterable[int | float | None]) -> float:
    """
    Get `nan`.

    :param data: ignored
    """
    if not isinstance(data, Iterable):
        raise type_error(data, "data", Iterable)
    return nan


def default_column_best(col: str) ->\
        Callable[[Iterable[int | float | None]], int | float]:
    """
    Get a function to compute the best value in a column.

    The returned function can compute the best value in a column. If no value
    is best, it should return `nan`.

    :param col: the column name string
    :returns: a function that can compute the best value per column
    """
    if not isinstance(col, str):
        raise type_error(col, "column name", str)

    if col in (KEY_INSTANCE, KEY_ALGORITHM, __KEY_LOWER_BOUND,
               __KEY_LOWER_BOUND_SHORT, KEY_GOAL_F):
        return __nan

    if SCOPE_SEPARATOR not in col:
        raise ValueError(
            f"statistic {col!r} should contain {SCOPE_SEPARATOR!r}.")
    key, stat = col.split(SCOPE_SEPARATOR)
    if (len(key) <= 0) or (len(stat) <= 0):
        raise ValueError(f"invalid statistic {col!r}.")

    if stat == "sd":
        return __finite_min

    if key in (KEY_BEST_F_SCALED, F_NAME_SCALED, KEY_BEST_F, F_NAME_RAW):
        return __finite_min
    if key in (KEY_LAST_IMPROVEMENT_TIME_MILLIS, KEY_LAST_IMPROVEMENT_FE,
               KEY_TOTAL_FES, KEY_TOTAL_TIME_MILLIS):
        return __finite_max

    return __nan


#: the number renderer for times
__TIME_NUMBER_RENDERER: Final[NumberRenderer] = \
    DEFAULT_NUMBER_RENDERER.derive(
        int_to_float_threshold=999_999,
        get_float_format=lambda _, ma, __, ___, itft:
        "{:.0f}" if ma <= itft else "{:.1e}")


def default_number_renderer(col: str) -> NumberRenderer:
    """
    Get the number renderer for the specified column.

    Time columns are rendered with less precision.

    :param col: the column name
    :returns: the number renderer
    """
    if not isinstance(col, str):
        raise type_error(col, "column name", str)
    if col not in (KEY_INSTANCE, KEY_ALGORITHM, __KEY_LOWER_BOUND,
                   __KEY_LOWER_BOUND_SHORT, KEY_GOAL_F):
        if SCOPE_SEPARATOR not in col:
            raise ValueError(
                f"statistic {col!r} should contain {SCOPE_SEPARATOR!r}.")
        key, stat = col.split(SCOPE_SEPARATOR)
        if (len(key) <= 0) or (len(stat) <= 0):
            raise ValueError(f"invalid statistic {col!r}.")
        if key in (KEY_LAST_IMPROVEMENT_TIME_MILLIS, KEY_LAST_IMPROVEMENT_FE,
                   KEY_TOTAL_FES, KEY_TOTAL_TIME_MILLIS):
            return __TIME_NUMBER_RENDERER
    return DEFAULT_NUMBER_RENDERER


def __getter(s: str) -> Callable[[EndStatistics], int | float | None]:
    """
    Obtain a getter for the end statistics.

    :param s: the name
    :returns: the getter
    """
    if not isinstance(s, str):
        raise type_error(s, "getter name", str)
    getter = EndStatistics.getter(s)

    def __fixed(e: EndStatistics, g=getter, n=s) -> int | float | None:
        res = g(e)
        if res is None:
            return None
        if not isinstance(res, int | float):
            raise type_error(res, f"result of getter {n!r} for statistic {e}",
                             (int, float))
        return res
    return __fixed


#: the default algorithm-instance statistics
DEFAULT_ALGORITHM_INSTANCE_STATISTICS: Final[tuple[
    str, str, str, str, str, str]] = (
    f"{KEY_BEST_F}{SCOPE_SEPARATOR}{KEY_MINIMUM}",
    f"{KEY_BEST_F}{SCOPE_SEPARATOR}{KEY_MEAN_ARITH}",
    f"{KEY_BEST_F}{SCOPE_SEPARATOR}{KEY_STDDEV}",
    f"{KEY_BEST_F_SCALED}{SCOPE_SEPARATOR}{KEY_MEAN_ARITH}",
    f"{KEY_LAST_IMPROVEMENT_FE}{SCOPE_SEPARATOR}{KEY_MEAN_ARITH}",
    f"{KEY_LAST_IMPROVEMENT_TIME_MILLIS}{SCOPE_SEPARATOR}"
    f"{KEY_MEAN_ARITH}")

#: the default algorithm summary statistics
DEFAULT_ALGORITHM_SUMMARY_STATISTICS: Final[tuple[
    str, str, str, str, str, str]] = (
    f"{KEY_BEST_F_SCALED}{SCOPE_SEPARATOR}{KEY_MINIMUM}",
    f"{KEY_BEST_F_SCALED}{SCOPE_SEPARATOR}{KEY_MEAN_GEOM}",
    f"{KEY_BEST_F_SCALED}{SCOPE_SEPARATOR}{KEY_MAXIMUM}",
    f"{KEY_BEST_F_SCALED}{SCOPE_SEPARATOR}{KEY_STDDEV}",
    f"{KEY_LAST_IMPROVEMENT_FE}{SCOPE_SEPARATOR}{KEY_MEAN_ARITH}",
    f"{KEY_LAST_IMPROVEMENT_TIME_MILLIS}{SCOPE_SEPARATOR}"
    f"{KEY_MEAN_ARITH}")


def tabulate_end_results(
        end_results: Iterable[EndResult],
        file_name: str = "table",
        dir_name: str = ".",
        algorithm_instance_statistics: Iterable[str] =
        DEFAULT_ALGORITHM_INSTANCE_STATISTICS,
        algorithm_summary_statistics: Iterable[str | None] | None =
        DEFAULT_ALGORITHM_SUMMARY_STATISTICS,
        text_format_driver: TextFormatDriver | Callable[[], TextFormatDriver]
        = Markdown.instance,
        algorithm_sort_key: Callable[[str], Any] = lambda a: a,
        instance_sort_key: Callable[[str], Any] = lambda i: i,
        col_namer: Callable[[str], str] = default_column_namer,
        col_best: Callable[[str], Callable[
            [Iterable[int | float | None]],
            int | float]] = default_column_best,
        col_renderer: Callable[[str], NumberRenderer] =
        default_number_renderer,
        put_lower_bound: bool = True,
        lower_bound_getter: Callable[[EndStatistics],
                                     int | float | None] | None =
        __getter(KEY_GOAL_F),
        lower_bound_name: str | None = __KEY_LOWER_BOUND,
        use_lang: bool = True,
        instance_namer: Callable[[str], str] = lambda x: x,
        algorithm_namer: Callable[[str], str] = lambda x: x) -> Path:
    r"""
    Tabulate the statistics about the end results of an experiment.

    A two-part table is produced. In the first part, it presents summary
    statistics about each instance-algorithm combination, sorted by instance.
    In the second part, it presents summary statistics of the algorithms over
    all instances. The following default columns are provided:

    1. Part 1: Algorithm-Instance statistics
        - `I`: the instance name
        - `lb(f)`: the lower bound of the objective value of
          the instance
        - `setup`: the name of the algorithm or algorithm setup
        - `best`: the best objective value reached by any run on that instance
        - `mean`: the arithmetic mean of the best objective values reached
          over all runs
        - `sd`: the standard deviation of the best objective values reached
          over all runs
        - `mean1`: the arithmetic mean of the best objective values reached
          over all runs, divided by the lower bound (or goal objective value)
        - `mean(FE/ms)`: the arithmetic mean of objective function evaluations
           performed per millisecond, over all runs
        - `mean(t)`: the arithmetic mean of the time in milliseconds when the
          last improving move of a run was applied, over all runs

    2. Part 2: Algorithm Summary Statistics
        - `setup`: the name of the algorithm or algorithm setup
        - `best1`: the minimum of the best objective values reached divided by
          the lower bound (or goal objective value) over all runs
        - `gmean1`: the geometric mean of the best objective values reached
           divided by the lower bound (or goal objective value) over all runs
        - `worst1`: the maximum of the best objective values reached divided
          by the lower bound (or goal objective value) over all runs
        - `sd1`: the standard deviation of the best objective values reached
           divided by the lower bound (or goal objective value) over all runs
        - `gmean(FE/ms)`: the geometric mean of objective function evaluations
          performed per millisecond, over all runs
        - `gmean(t)`: the geometric mean of the time in milliseconds when the
          last improving move of a run was applied, over all runs

    You can freely configure which columns you want for each part and whether
    you want to have the second part included. Also, for each group of values,
    the best one is marked in bold face.

    Depending on the parameter `text_format_driver`, the tables can be
    rendered in different formats, such as
    :py:class:`~moptipy.utils.markdown.Markdown`,
    :py:class:`~moptipy.utils.latex.LaTeX`, and
    :py:class:`~moptipy.utils.html.HTML`.

    :param end_results: the end results data
    :param file_name: the base file name
    :param dir_name: the base directory
    :param algorithm_instance_statistics: the statistics to print
    :param algorithm_summary_statistics: the summary statistics to print per
        algorithm
    :param text_format_driver: the text format driver
    :param algorithm_sort_key: a function returning sort keys for algorithms
    :param instance_sort_key: a function returning sort keys for instances
    :param col_namer: the column namer function
    :param col_best: the column-best getter function
    :param col_renderer: the number renderer for the column
    :param put_lower_bound: should we put the lower bound or goal objective
        value?
    :param lower_bound_getter: the getter for the lower bound
    :param lower_bound_name: the name key for the lower bound to be passed
        to `col_namer`
    :param use_lang: should we use the language to define the filename?
    :param instance_namer: the name function for instances receives an
        instance ID and returns an instance name; default=identity function
    :param algorithm_namer: the name function for algorithms receives an
        algorithm ID and returns an instance name; default=identity function
    :returns: the path to the file with the tabulated end results
    """
    # Before doing anything, let's do some type checking on the parameters.
    # I want to ensure that this function is called correctly before we begin
    # to actually process the data. It is better to fail early than to deliver
    # some incorrect results.
    if not isinstance(end_results, Iterable):
        raise type_error(end_results, "end_results", Iterable)
    if not isinstance(file_name, str):
        raise type_error(file_name, "file_name", str)
    if not isinstance(dir_name, str):
        raise type_error(dir_name, "dir_name", str)
    if not isinstance(algorithm_instance_statistics, Iterable):
        raise type_error(algorithm_instance_statistics,
                         "algorithm_instance_statistics", Iterable)
    if (algorithm_summary_statistics is not None)\
            and (not isinstance(algorithm_instance_statistics, Iterable)):
        raise type_error(algorithm_summary_statistics,
                         "algorithm_summary_statistics", (Iterable, None))
    if not isinstance(put_lower_bound, bool):
        raise type_error(put_lower_bound, "put_lower_bound", bool)
    if put_lower_bound:
        if not callable(lower_bound_getter):
            raise type_error(lower_bound_getter, "lower_bound_getter",
                             call=True)
        if not isinstance(lower_bound_name, str):
            raise type_error(lower_bound_name, "lower_bound_name", str)
    if callable(text_format_driver):
        text_format_driver = text_format_driver()
    if not isinstance(text_format_driver, TextFormatDriver):
        raise type_error(text_format_driver, "text_format_driver",
                         TextFormatDriver, True)
    if not callable(col_namer):
        raise type_error(col_namer, "col_namer", call=True)
    if not callable(col_best):
        raise type_error(col_best, "col_best", call=True)
    if not callable(col_renderer):
        raise type_error(col_renderer, "col_renderer", call=True)
    if not isinstance(use_lang, bool):
        raise type_error(use_lang, "use_lang", bool)
    if not callable(algorithm_namer):
        raise type_error(algorithm_namer, "algorithm_namer", call=True)
    if not callable(instance_namer):
        raise type_error(instance_namer, "instance_namer", call=True)

    # quick protection of renderer
    def __col_renderer(col: str, __fwd=col_renderer) -> NumberRenderer:
        res = __fwd(col)
        if not isinstance(res, NumberRenderer):
            raise type_error(res, f"col_renderer({col!r})", NumberRenderer)
        return res

    # get the getters
    algo_inst_getters: Final[list[Callable[[EndStatistics],
                                           int | float | None]]] = \
        [__getter(d) for d in algorithm_instance_statistics]
    n_algo_inst_getters: Final[int] = len(algo_inst_getters)
    if n_algo_inst_getters <= 0:
        raise ValueError("algorithm-instance dimensions must not be empty.")
    algo_getters: Final[list[Callable[[EndStatistics],
                                      int | float | None] | None] | None] = \
        (None if (algorithm_summary_statistics is None)
         else [None if (d is None) else __getter(d)
               for d in cast(Iterable, algorithm_summary_statistics)])
    if algo_getters is not None:
        if len(algo_getters) != n_algo_inst_getters:
            raise ValueError(
                f"there are {n_algo_inst_getters} algorithm-instance columns,"
                f" but {len(algo_getters)} algorithms summary columns.")
        if all(g is None for g in algo_getters):
            raise ValueError(
                "if all elements of algorithm_summary_statistics are None, "
                "then specify algorithm_summary_statistics=None")

    # gather the statistics for each algorithm-instance combination
    algo_inst_list: Final[list[EndStatistics]] = []
    EndStatistics.from_end_results(end_results, algo_inst_list.append)
    if len(algo_inst_list) <= 0:
        raise ValueError("no algorithm-instance combinations?")
    # get the sorted lists of algorithms and instances
    insts: Final[list[str]] = sorted({s.instance for s in algo_inst_list},
                                     key=instance_sort_key)
    n_insts: Final[int] = len(insts)
    if n_insts <= 0:
        raise ValueError("no instance found?")
    inst_names: Final[list[str]] = [instance_namer(inst) for inst in insts]
    algos: Final[list[str]] = sorted({s.algorithm for s in algo_inst_list},
                                     key=algorithm_sort_key)
    n_algos: Final[int] = len(algos)
    if n_algos <= 0:
        raise ValueError("no algos found?")
    algo_names: Final[list[str]] = [algorithm_namer(algo) for algo in algos]

    # finalize the data dictionaries: d[inst][algo] = stats
    algo_inst_dict: Final[dict[str, dict[str, EndStatistics]]] = {}
    for e in algo_inst_list:
        if e.instance not in algo_inst_dict:
            algo_inst_dict[e.instance] = {}
        algo_inst_dict[e.instance][e.algorithm] = e
    for ina in insts:
        if len(algo_inst_dict[ina]) != n_algos:
            raise ValueError(
                f"expected {n_algos} entries for instance {ina!r}, but "
                f"got only {len(algo_inst_dict[ina])}, namely "
                f"{algo_inst_dict[ina].keys()} instead of {algos}.")

    # compute the per-instance lower bounds if we need them
    lower_bounds: list[str] | None
    if put_lower_bound:
        lb: list[int | float | None] = []
        for inst in insts:
            bounds = list({lower_bound_getter(d)
                           for d in algo_inst_dict[inst].values()})
            if len(bounds) != 1:
                raise ValueError(f"inconsistent lower bounds {bounds} for "
                                 f"instance {inst!r}.")
            lb.append(bounds[0])
        lower_bounds = cast(list[str], __col_renderer(
            __KEY_LOWER_BOUND).render(lb))
        del lb
    else:
        lower_bounds = None
    del algo_inst_list

    # gather the algorithm summary statistics
    algo_dict: Final[dict[str, EndStatistics] | None] = {} \
        if (n_insts > 1) and (algo_getters is not None) else None
    if algo_dict is not None:
        def __put(es: EndStatistics) -> None:
            nonlocal algo_dict
            if es.algorithm in algo_dict:
                raise ValueError(f"already encountered {es.algorithm}?")
            algo_dict[es.algorithm] = es
        EndStatistics.from_end_results(end_results, __put,
                                       join_all_instances=True)
        del __put
    del end_results
    if len(algo_dict) != n_algos:
        raise ValueError(f"there are {n_algos} algorithms, but in the "
                         f"summary, only {len(algo_dict)} appear?")

    # set up column titles
    def __fix_name(s: str, nn=col_namer) -> str:
        if not isinstance(s, str):
            raise type_error(s, "column name", str)
        if len(s) <= 0:
            raise ValueError("string must not be empty!")
        na = nn(s)
        if not isinstance(na, str):
            raise type_error(na, f"name computed for {s}", str)
        if len(na) <= 0:
            raise ValueError(f"name computed for {s} cannot be empty.")
        return na

    algo_inst_cols: Final[list[str]] = \
        [__fix_name(s) for s in algorithm_instance_statistics]
    if len(algo_inst_cols) <= 0:
        raise ValueError("no algorithm_instance columns?")
    algo_cols: list[str | None] | None = \
        None if algo_dict is None else \
        [(None if s is None else __fix_name(s))
         for s in cast(Iterable, algorithm_summary_statistics)]
    if algo_cols == algo_inst_cols:
        algo_cols = None  # no need to repeat header if it is the same

    # set up the column definitions
    algo_inst_cols.insert(0, __fix_name(KEY_ALGORITHM))
    if put_lower_bound:
        algo_inst_cols.insert(0, __fix_name(lower_bound_name))
    algo_inst_cols.insert(0, __fix_name(KEY_INSTANCE))

    if algo_cols is not None:
        algo_cols.insert(0, __fix_name(KEY_ALGORITHM))
        if put_lower_bound:
            algo_cols.insert(0, None)
        algo_cols.insert(0, None)

    col_def: Final[str] = ("lrl" if put_lower_bound else "ll") \
        + ("r" * n_algo_inst_getters)

    # get the data columns of all columns and convert to strings
    # format: column -> columns data
    # we first need to get all the data at once to allow for a uniform
    # formatting via numbers_to_strings
    algo_inst_data_raw: Final[list[list[int | float | None]]] =\
        [[None if getter is None else getter(algo_inst_dict[inst][algo])
          for inst in insts for algo in algos]
         for getter in algo_inst_getters]
    algo_inst_strs_raw: Final[list[list[str | None]]] = [
        cast(list[str | None],
             __col_renderer(ais).render(algo_inst_data_raw[i]))
        for i, ais in enumerate(algorithm_instance_statistics)]

    # now break the data into sections
    # format: column -> per-instance section -> section data
    # after we break the data in sections, we can mark the per-section bests
    # and we can flush the data to the table section-wise
    algo_inst_data: Final[list[list[list[int | float | None]]]] = \
        [[col[c * n_algos:(c + 1) * n_algos] for c in range(n_insts)]
         for col in algo_inst_data_raw]
    del algo_inst_data_raw
    algo_inst_strs: Final[list[list[list[str | None]]]] = \
        [[col[c * n_algos:(c + 1) * n_algos] for c in range(n_insts)]
         for col in algo_inst_strs_raw]
    del algo_inst_strs_raw

    # now format the data, i.e., compute the per-section best value
    # of each column and mark it with bold face
    for col_idx, stat in enumerate(algorithm_instance_statistics):
        col_n = algo_inst_data[col_idx]
        col_s = algo_inst_strs[col_idx]
        best_getter = col_best(stat)
        if not callable(best_getter):
            raise type_error(best_getter, f"result of col_best for {stat}",
                             call=True)
        for chunk_idx, chunk_n in enumerate(col_n):
            chunk_s = col_s[chunk_idx]
            best = best_getter(chunk_n)
            if (best is not None) and (not isnan(best)):
                for idx, val in enumerate(chunk_n):
                    if val == best:
                        chunk_s[idx] = FormattedStr.add_format(
                            chunk_s[idx], bold=True)
    del algo_inst_data
    del algorithm_instance_statistics

    # now we pre-pend the instance and algorithm information
    algo_names_formatted: list[str] = [
        FormattedStr.add_format(algo, code=True) for algo in algo_names]
    algo_inst_strs.insert(0, [algo_names_formatted] * n_insts)
    if put_lower_bound:
        algo_inst_strs.insert(0, [[b] for b in lower_bounds])
    algo_inst_strs.insert(0, [[
        FormattedStr.add_format(inst, code=True)] for inst in inst_names])
    del lower_bounds
    del insts

    algo_strs: list[list[str | None]] | None = None
    if (algo_dict is not None) and (algorithm_summary_statistics is not None):
        # get the data columns of the algorithm summaries
        # format: column -> columns data
        algo_data: Final[list[list[int | float | None]]] = \
            [[None if getter is None else getter(algo_dict[algo])
              for algo in algos]
             for getter in algo_getters]
        algo_strs = [cast(list[str], __col_renderer(ass).render(algo_data[i]))
                     for i, ass in enumerate(algorithm_summary_statistics)]

        # now format the data, i.e., compute the per-section best value
        # of each column and mark it with bold face
        for col_idx, stat in enumerate(algorithm_summary_statistics):
            if stat is None:
                continue
            acol_n = algo_data[col_idx]
            acol_s = algo_strs[col_idx]
            best_getter = col_best(stat)
            if not callable(best_getter):
                raise type_error(
                    best_getter, f"result of col_best for {stat}", call=True)
            best = best_getter(acol_n)
            if (best is not None) and (not isnan(best)):
                for idx, val in enumerate(acol_n):
                    if val == best:
                        acol_s[idx] = FormattedStr.add_format(
                            acol_s[idx], bold=True)
        del algo_data
        algo_strs.insert(0, algo_names_formatted)
        if put_lower_bound:
            algo_strs.insert(0, [])
        algo_strs.insert(0, [__fix_name("summary")] * n_algos)

    # write the table
    dest: Final[Path] = text_format_driver.filename(
        file_name, dir_name, use_lang)
    with dest.open_for_write() as wd, \
            Table(wd, col_def, text_format_driver) as table:
        with table.header() as head:
            head.full_row(algo_inst_cols)
        for i in range(n_insts):
            with table.section() as sec:
                sec.cols([col[i] for col in algo_inst_strs])
        if algo_strs is not None:
            with table.section() as sec:
                if algo_cols is not None:
                    with sec.header() as head:
                        head.full_row(algo_cols)
                sec.cols(algo_strs)

    return dest



"""Styler allows to discover groups of data and associate styles with them."""
from typing import Any, Callable, Final, Iterable, cast

from matplotlib.artist import Artist  # type: ignore
from matplotlib.lines import Line2D  # type: ignore

from moptipy.utils.plot_defaults import create_line_style
from moptipy.utils.types import type_error


class Styler:
    """A class for determining groups of elements and styling them."""

    #: The tuple with the names becomes valid after compilation.
    names: tuple[str, ...]
    #: The tuple with the keys becomes valid after compilation.
    keys: tuple[Any, ...]
    #: The dictionary mapping keys to indices; only valid after compilation.
    __indexes: dict[Any, int]
    #: Is there a None key? Valid after compilation.
    has_none: bool
    #: The number of registered keys.
    count: int
    #: Does this styler have any style associated with it?
    has_style: bool

    def __init__(self,
                 key_func: Callable = lambda x: x,
                 namer: Callable[[Any], str] = str,
                 none_name: str = "None",
                 priority: int | float = 0,
                 name_sort_function: Callable[[str], Any] | None =
                 lambda s: s):
        """
        Initialize the style grouper.

        :param key_func: the key function, obtaining keys from objects
        :param namer: the name function, turning keys into names
        :param none_name: the name for the none-key
        :param priority: the base priority of this grouper
        :param name_sort_function: the function for sorting names, or `None`
            if no name-based sorting shall be performed
        """
        if not callable(key_func):
            raise type_error(key_func, "key function", call=True)

        if not callable(namer):
            raise type_error(namer, "namer function", call=True)
        if not isinstance(none_name, str):
            raise type_error(none_name, "none_name", str)
        if not isinstance(priority, float | int):
            raise type_error(priority, "priority", (int, float))
        if (name_sort_function is not None) \
                and (not callable(name_sort_function)):
            raise type_error(name_sort_function, "name_sort_function",
                             type(None), call=True)

        def __namer(key,
                    __namer: Callable[[Any], str] = namer,
                    __none_name: str = none_name) -> str:
            rv = __none_name if key is None else __namer(key)
            if not isinstance(rv, str):
                raise type_error(rv, f"name for key {key!r}", str)
            rv = rv.strip()
            if len(rv) <= 0:
                raise ValueError(
                    "name cannot be empty or just consist of white space")
            return rv

        #: the name sort function
        self.__name_sort_function: Final[Callable[[str], Any] | None] = \
            name_sort_function
        #: The key function of the grouper
        self.key_func: Final[Callable] = key_func
        #: The name function of the grouper
        self.name_func: Final[Callable[[Any], str]] = \
            cast(Callable[[Any], str], __namer)
        #: The base priority of this grouper
        self.priority: float = float(priority)
        #: The internal collection.
        self.__collection: set = set()
        #: the line colors
        self.__line_colors: tuple | None = None
        #: the line dashes
        self.__line_dashes: tuple | None = None
        #: the line widths
        self.__line_widths: tuple[float, ...] | None = None
        #: the optional line alpha
        self.__line_alphas: tuple[float, ...] | None = None

    def add(self, obj) -> None:
        """
        Add an object to the style collection.

        :param obj: the object
        """
        self.__collection.add(self.key_func(obj))

    def finalize(self) -> None:
        """Compile the styler collection."""
        self.has_none = (None in self.__collection)
        if self.has_none:
            self.__collection.remove(None)

        nsf: Final[Callable[[str], Any] | None] = self.__name_sort_function
        if nsf is None:
            data = [(k, self.name_func(k)) for k in self.__collection]
            data.sort()
            if self.has_none:
                data.insert(0, (None, self.name_func(None)))
            self.keys = tuple([x[0] for x in data])
            self.names = tuple([x[1] for x in data])
        else:
            data = [(self.name_func(k), k) for k in self.__collection]
            data.sort(key=cast(Callable, lambda x, nsf2=nsf: nsf2(x[0])))
            if self.has_none:
                data.insert(0, (self.name_func(None), None))
            self.names = tuple([x[0] for x in data])
            self.keys = tuple([x[1] for x in data])

        del self.__collection
        del data
        self.__indexes = {k: i for i, k in enumerate(self.keys)}
        self.count = len(self.names)
        self.priority += self.count
        self.has_style = False

    def __lt__(self, other) -> bool:
        """
        Check whether this styler is more important than another one.

        :param other: the other styler
        :return: `True` if it is, `False` if it is not.
        """
        if self.priority > other.priority:
            return True
        if self.priority < other.priority:
            return False
        c1 = self.count
        if self.has_none:
            c1 -= 1
        c2 = other.count
        if other.has_none:
            c2 -= 1
        if c1 > c2:
            return True
        return False

    def set_line_color(self, line_color_func: Callable) -> None:
        """
        Set that this styler should apply line colors.

        :param line_color_func: a function returning the palette
        """
        tmp = line_color_func(self.count)
        if not isinstance(tmp, Iterable):
            raise type_error(tmp, "result of line color func", Iterable)
        self.__line_colors = tuple(tmp)
        if len(self.__line_colors) != self.count:
            raise ValueError(f"There must be {self.count} line colors,"
                             f"but found only {len(self.__line_colors)}.")
        self.has_style = True

    def set_line_dash(self, line_dash_func: Callable) -> None:
        """
        Set that this styler should apply line dashes.

        :param line_dash_func: a function returning the dashes
        """
        tmp = line_dash_func(self.count)
        if not isinstance(tmp, Iterable):
            raise type_error(tmp, "result of line dash func", Iterable)
        self.__line_dashes = tuple(tmp)
        if len(self.__line_dashes) != self.count:
            raise ValueError(f"There must be {self.count} line dashes,"
                             f"but found only {len(self.__line_dashes)}.")
        self.has_style = True

    def set_line_width(self, line_width_func: Callable) -> None:
        """
        Set that this styler should apply a line width.

        :param line_width_func: the line width function
        """
        tmp = line_width_func(self.count)
        if not isinstance(tmp, Iterable):
            raise type_error(tmp, "result of line width func", Iterable)
        self.__line_widths = tuple(tmp)
        if len(self.__line_widths) != self.count:
            raise ValueError(f"There must be {self.count} line widths,"
                             f"but found only {len(self.__line_widths)}.")
        self.has_style = True

    def set_line_alpha(self, line_alpha_func: Callable) -> None:
        """
        Set that this styler should apply a line alpha.

        :param line_alpha_func: the line alpha function
        """
        tmp = line_alpha_func(self.count)
        if not isinstance(tmp, Iterable):
            raise type_error(tmp, "result of line alpha func", Iterable)
        self.__line_alphas = tuple(tmp)
        if len(self.__line_alphas) != self.count:
            raise ValueError(f"There must be {self.count} line alphas,"
                             f"but found only {len(self.__line_alphas)}.")
        self.has_style = True

    def add_line_style(self, obj,
                       style: dict[str, object]) -> None:
        """
        Apply this styler's contents based on the given object.

        :param obj: the object for which the style should be created
        :param style: the decode to which the styles should be added
        """
        key = self.key_func(obj)
        index = self.__indexes.setdefault(key, -1)
        if index >= 0:
            self.__add_line_style(index, style)

    def __add_line_style(self, index,
                         style: dict[str, object]) -> None:
        """
        Apply this styler's contents based on the given object.

        :param index: the index to be processed
        :param style: the decode to which the styles should be added
        """
        if self.__line_colors is not None:
            style["color"] = self.__line_colors[index]
        if self.__line_dashes is not None:
            style["linestyle"] = self.__line_dashes[index]
        if self.__line_widths is not None:
            style["linewidth"] = self.__line_widths[index]
        if self.__line_alphas is not None:
            style["alpha"] = self.__line_alphas[index]

    def add_to_legend(self, consumer: Callable[[Artist], Any]) -> None:
        """
        Add this styler to the legend.

        :param consumer: the consumer to add to
        """
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        for i, name in enumerate(self.names):
            style = create_line_style()
            self.__add_line_style(i, style)
            style["label"] = name
            style["xdata"] = []
            style["ydata"] = []
            consumer(Line2D(**style))



"""An internal module with the base class for implementing Processes."""
import os
from io import StringIO
from math import inf, isfinite
from threading import Lock, Timer
from time import time_ns
from traceback import print_tb
from typing import Any, Callable, Final, cast

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm, check_algorithm
from moptipy.api.logging import (
    _ALL_SECTIONS,
    KEY_BEST_F,
    KEY_EXCEPTION_STACK_TRACE,
    KEY_EXCEPTION_TYPE,
    KEY_EXCEPTION_VALUE,
    KEY_GOAL_F,
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_MAX_FES,
    KEY_MAX_TIME_MILLIS,
    KEY_RAND_BIT_GENERATOR_TYPE,
    KEY_RAND_GENERATOR_TYPE,
    KEY_RAND_SEED,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
    SCOPE_ALGORITHM,
    SCOPE_OBJECTIVE_FUNCTION,
    SCOPE_PROCESS,
    SCOPE_SOLUTION_SPACE,
    SECTION_ERROR_BEST_F,
    SECTION_ERROR_IN_CONTEXT,
    SECTION_ERROR_IN_LOG,
    SECTION_ERROR_IN_RUN,
    SECTION_ERROR_INVALID_X,
    SECTION_ERROR_INVALID_Y,
    SECTION_ERROR_TIMING,
    SECTION_FINAL_STATE,
    SECTION_RESULT_Y,
    SECTION_SETUP,
)
from moptipy.api.objective import Objective, check_objective
from moptipy.api.process import (
    Process,
    check_goal_f,
    check_max_fes,
    check_max_time_millis,
)
from moptipy.api.space import Space, check_space
from moptipy.utils.logger import (
    SECTION_END,
    SECTION_START,
    FileLogger,
    KeyValueLogSection,
    Logger,
)
from moptipy.utils.nputils import (
    rand_generator,
    rand_seed_check,
    rand_seed_generate,
)
from moptipy.utils.path import Path
from moptipy.utils.sys_info import log_sys_info
from moptipy.utils.types import type_error, type_name_of


def _error_1(logger: Logger, title: str, exception_type,
             exception_value, traceback) -> None:
    """
    Create a text section with error information as from a contextmanager.

    :param logger: the logger to write to
    :param title: the title of the section with error information to be
        created
    :param exception_type: the exception type
    :param exception_value: the exception value
    :param traceback: the traceback
    """
    if exception_type or exception_value or traceback:
        with logger.text(title=title) as ts:
            if exception_type:
                ts.write(KEY_EXCEPTION_TYPE)
                ts.write(": ")
                if isinstance(exception_type, str):
                    if exception_type.startswith("<class '"):
                        exception_type = exception_type[8:-2]
                    ts.write(exception_type.strip())
                else:
                    ts.write(type_name_of(exception_type))
                ts.write(os.linesep)
            if exception_value:
                ts.write(KEY_EXCEPTION_VALUE)
                ts.write(": ")
                ts.write(str(exception_value).strip())
                ts.write(os.linesep)
            if traceback:
                ts.write(KEY_EXCEPTION_STACK_TRACE)
                ts.write(":")
                ts.write(os.linesep)
                sio = StringIO()
                print_tb(traceback, file=sio)
                for line in sio.getvalue().split("\n"):
                    ts.write(line.strip())
                    ts.write(os.linesep)


def _error_2(logger: Logger, title: str, exception: BaseException) -> None:
    """
    Log an exception.

    :param logger: the logger to write to
    :param title: the title of the section with error information to be
        created
    :param exception: the exception

    >>> from moptipy.utils.logger import InMemoryLogger
    >>> ime = InMemoryLogger()
    >>> def k():
    ...     1 / 0
    >>> try:
    ...     k()
    ... except BaseException as be:
    ...     _error_2(ime, "ERROR", be)
    >>> print(ime.get_log()[0])
    BEGIN_ERROR
    >>> print(ime.get_log()[1])
    exceptionType: ZeroDivisionError
    >>> print(ime.get_log()[2])
    exceptionValue: division by zero
    >>> print(ime.get_log()[3])
    exceptionStackTrace:
    >>> print(ime.get_log()[-1])
    END_ERROR
    """
    _error_1(logger, title, exception_type=exception,
             exception_value=str(exception),
             traceback=exception.__traceback__)


# the function used to get the time
_TIME_IN_NS: Final[Callable[[], int]] = time_ns


def _ns_to_ms(nanos: int) -> int:
    """
    Convert nanoseconds to milliseconds by rounding up.

    :param int nanos: the nanoseconds
    :returns: the corresponding milliseconds, rounded up

    >>> _ns_to_ms(0)
    0
    >>> _ns_to_ms(1)
    1
    >>> _ns_to_ms(999_999)
    1
    >>> _ns_to_ms(1_000_000)
    1
    >>> _ns_to_ms(1_000_001)
    2
    >>> _ns_to_ms(1_999_999)
    2
    >>> _ns_to_ms(2_000_000)
    2
    >>> _ns_to_ms(2_000_001)
    3
    """
    return (nanos + 999_999) // 1_000_000


class _ProcessBase(Process):
    """The internal base class for implementing optimization processes."""

    def __init__(self,
                 solution_space: Space,
                 objective: Objective,
                 algorithm: Algorithm,
                 log_file: Path | None = None,
                 rand_seed: int | None = None,
                 max_fes: int | None = None,
                 max_time_millis: int | None = None,
                 goal_f: int | float | None = None) -> None:
        """
        Perform the internal initialization. Do not call directly.

        :param solution_space: the search- and solution space.
        :param objective: the objective function
        :param algorithm: the optimization algorithm
        :param log_file: the optional log file
        :param rand_seed: the optional random seed
        :param max_fes: the maximum permitted function evaluations
        :param max_time_millis: the maximum runtime in milliseconds
        :param goal_f: the goal objective value. if it is reached, the process
            is terminated
        """
        super().__init__()
        #: This will be `True` after :meth:`terminate` has been called.
        self._terminated: bool = False
        #: This becomes `True` when :meth:`should_terminate` returned `True`.
        self._knows_that_terminated: bool = False
        #: The internal lock, needed to protect :meth:`terminate`.
        self.__lock: Final[Lock] = Lock()
        #: The maximum FEs.
        self._max_fes: Final[int | None] = check_max_fes(max_fes, True)
        #: A version of :attr:`_max_fes` that can be used in comparisons.
        self._end_fes: Final[int] = 9_223_372_036_854_775_800 \
            if (self._max_fes is None) else self._max_fes
        #: The goal objective value.
        self._goal_f: Final[int | float | None] = \
            check_goal_f(goal_f, True)
        #: A comparable version of :attr:`self._goal_f`.
        self._end_f: int | float = \
            -inf if (self._goal_f is None) else self._goal_f
        #: The currently consumed nanoseconds.
        self._current_time_nanos: int = 0
        #: The currently consumed objective function evaluations (FEs).
        self._current_fes: int = 0
        #: The time (in nanoseconds) when the last improvement was made.
        self._last_improvement_time_nanos: int = -1
        #: The FE when the last improvement was made.
        self._last_improvement_fe: int = -1

        #: The solution space, i.e., the data structure of possible solutions.
        self._solution_space: Final[Space] = check_space(solution_space)
        #: The objective function rating candidate solutions.
        self.__objective: Final[Objective] = check_objective(objective)
        #: the internal invoker for the objective function
        self._f: Final[Callable[[Any], int | float]] = \
            self.__objective.evaluate
        #: The algorithm to be applied.
        self.__algorithm: Final[Algorithm] = check_algorithm(algorithm)
        #: The random seed.
        self.__rand_seed: Final[int] = rand_seed_generate() \
            if rand_seed is None \
            else rand_seed_check(rand_seed)
        #: The random number generator.
        self.__random: Final[Generator] = rand_generator(self.__rand_seed)
        #: The current best solution.
        self._current_best_y = solution_space.create()
        #: The current best objective value
        self._current_best_f: int | float = inf
        #: The log file, or `None` is needed
        if (log_file is not None) and (not isinstance(log_file, Path)):
            raise type_error(log_file, "log_file", Path)
        self.__log_file: Final[Path | None] = log_file
        #: the method for copying y
        self._copy_y: Final[Callable] = solution_space.copy
        #: set up the method forwards
        self.lower_bound = self.__objective.lower_bound  # type: ignore
        if self._end_f <= -inf:
            self._end_f = check_goal_f(self.lower_bound())
            self.lower_bound = lambda: self._end_f  # type: ignore
        self.upper_bound = objective.upper_bound  # type: ignore
        self.is_always_integer = objective.is_always_integer  # type: ignore
        self.create = solution_space.create  # type: ignore
        self.copy = solution_space.copy  # type: ignore
        self.to_str = solution_space.to_str  # type: ignore
        self.is_equal = solution_space.is_equal  # type: ignore
        self.from_str = solution_space.from_str  # type: ignore
        self.n_points = solution_space.n_points  # type: ignore
        self.validate = solution_space.validate  # type: ignore
        #: the internal section logger
        self.__sections: dict[str, str] | None = \
            None if log_file is None else {}

        #: The time when the process was started, in nanoseconds.
        self._start_time_nanos: Final[int] = _TIME_IN_NS()
        #: The maximum runtime in milliseconds.
        self._max_time_millis: Final[int | None] = \
            check_max_time_millis(max_time_millis, True)
        #: A comparable version of :attr:`_max_time_millis`, but representing
        #: the end time in nanoseconds rounded to the next highest
        #: millisecond.
        self._end_time_nanos: Final[float | int] = \
            inf if (self._max_time_millis is None) else \
            _ns_to_ms(int(self._start_time_nanos
                          + (1_000_000 * self._max_time_millis))) \
            * 1_000_000
        #: The timer until the end-of-run, or `None` if there is no end time.
        self.__timer: Final[Timer | None] = None \
            if (self._max_time_millis is None) else \
            Timer(interval=self._max_time_millis / 1_000.0,
                  function=self.terminate)

        #: an internal base exception caught by the algorithm execution
        self._caught: BaseException | None = None

    def _after_init(self) -> None:
        """
        Finish initialization, start timer for termination if needed.

        Internal method that must be called after __init__ is completed.
        """
        if self.__timer is not None:
            self.__timer.start()

    def get_random(self) -> Generator:
        return self.__random

    def should_terminate(self) -> bool:
        if self._terminated:
            self._knows_that_terminated = True
            return True
        return False

    def get_consumed_fes(self) -> int:
        return self._current_fes

    def get_consumed_time_millis(self) -> int:
        if not self._terminated:
            self._current_time_nanos = time = _TIME_IN_NS()
            if time >= self._end_time_nanos:
                self.terminate()
        return _ns_to_ms(self._current_time_nanos - self._start_time_nanos)

    def get_max_time_millis(self) -> int | None:
        return self._max_time_millis

    def get_max_fes(self) -> int | None:
        return self._max_fes

    def get_last_improvement_fe(self) -> int:
        if self._last_improvement_fe < 0:
            raise ValueError("Did not perform FE yet, cannot query "
                             "last improvement FE.")
        return self._last_improvement_fe

    def get_last_improvement_time_millis(self) -> int:
        if self._last_improvement_time_nanos < 0:
            raise ValueError("Did not perform FE yet, cannot query "
                             "last improvement time.")
        return _ns_to_ms(self._last_improvement_time_nanos
                         - self._start_time_nanos)

    def has_best(self) -> bool:
        return self._current_fes > 0

    def get_best_f(self) -> int | float:
        if self._current_fes > 0:
            return self._current_best_f
        raise ValueError("No current best available.")

    def get_copy_of_best_x(self, x) -> None:
        if self._current_fes > 0:
            return self._copy_y(x, self._current_best_y)
        raise ValueError("No current best available.")

    def terminate(self) -> None:
        with self.__lock:
            old_terminated = self._terminated
            self._terminated = True
            if old_terminated:
                return
            if self.__timer is not None:
                self.__timer.cancel()
            del self.__timer
            self._current_time_nanos = _TIME_IN_NS()

    def get_copy_of_best_y(self, y) -> None:
        """
        Get a copy of the current best point in the solution space.

        This method in this internal class just forwards to
        :meth:`get_copy_of_best_x`.

        :param y: the destination data structure to be overwritten
        """
        return self.get_copy_of_best_x(y)

    def _log_own_parameters(self, logger: KeyValueLogSection) -> None:
        """
        Write the parameters of this process to the logger.

        This includes the limits on runtime and FEs.

        :param logger: the logger
        """
        super().log_parameters_to(logger)
        if self._max_fes is not None:
            logger.key_value(KEY_MAX_FES, self._max_fes)
        if self._max_time_millis is not None:
            logger.key_value(KEY_MAX_TIME_MILLIS, self._max_time_millis)
        if self._goal_f is not None:
            logger.key_value(KEY_GOAL_F, self._goal_f)
        logger.key_value(KEY_RAND_SEED, self.__rand_seed, True)
        logger.key_value(KEY_RAND_GENERATOR_TYPE, type_name_of(self.__random))
        logger.key_value(KEY_RAND_BIT_GENERATOR_TYPE,
                         type_name_of(self.__random.bit_generator))

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Write the standard parameters of this process to the logger.

        This includes the limits on runtime and FEs.

        :param logger: the logger
        """
        with logger.scope(SCOPE_PROCESS) as sc:
            self._log_own_parameters(sc)
        with logger.scope(SCOPE_ALGORITHM) as sc:
            self.__algorithm.log_parameters_to(sc)
        with logger.scope(SCOPE_SOLUTION_SPACE) as sc:
            self._solution_space.log_parameters_to(sc)
        with logger.scope(SCOPE_OBJECTIVE_FUNCTION) as sc:
            self.__objective.log_parameters_to(sc)

    def add_log_section(self, title: str, text: str) -> None:
        """
        Add a section to the log, if a log is written (otherwise ignore it).

        :param title: the title of the log section
        :param text: the text to log
        """
        if not isinstance(title, str):
            raise type_error(title, "title", str)
        t = title.strip()
        if (len(t) != len(title)) or (len(t) <= 0) or (" " in t) \
                or ("\n" in t) or ("\t" in t):
            raise ValueError("section title must not be empty or contain "
                             f"white space, but {title!r} is/does.")
        if (t in _ALL_SECTIONS) or (SECTION_START in t) or (SECTION_END in t):
            raise ValueError(f"title {t!r} is a reserved section title")
        if t.upper() != t:
            raise ValueError("section titles must be in upper case,"
                             f"but yours is {t!r} (vs. {t.upper()!r}.")
        for ch in t:  # check all character codes in t
            code: int = ord(ch)  # we will only permit A-Z, 0-9, and _
            if not ((65 <= code <= 90) or (48 <= code <= 57) or (code == 95)):
                raise ValueError(
                    f"{ch!r} forbidden in section title, but got {t!r}.")
        if not isinstance(text, str):
            raise type_error(text, "text", str)
        if (SECTION_START in text) or (SECTION_END in text):
            raise ValueError(
                f"text of section {t!r} must not contain {SECTION_START!r} or"
                f" {SECTION_END!r} but is {text!r}")
        if self.__sections is not None:
            if title in self.__sections:
                raise ValueError(f"section {title!r} already logged.")
            self.__sections[title] = text.strip()

    def _log_best(self, kv: KeyValueLogSection) -> None:
        """
        Log the best solution.

        :param kv: the key values logger
        """
        kv.key_value(KEY_BEST_F, self._current_best_f)
        kv.key_value(KEY_LAST_IMPROVEMENT_FE,
                     self._last_improvement_fe)
        kv.key_value(KEY_LAST_IMPROVEMENT_TIME_MILLIS,
                     _ns_to_ms(self._last_improvement_time_nanos
                               - self._start_time_nanos))

    def _write_result(self, logger: Logger) -> None:
        """
        Write the end result into the log.

        :param logger: the logger
        """
        with logger.text(SECTION_RESULT_Y) as txt:
            txt.write(self._solution_space.to_str(self._current_best_y))

    def _write_log(self, logger: Logger) -> None:
        """Write the information gathered during optimization into the log."""
        with logger.key_values(SECTION_FINAL_STATE) as kv:
            kv.key_value(KEY_TOTAL_FES, self._current_fes)
            kv.key_value(KEY_TOTAL_TIME_MILLIS,
                         _ns_to_ms(self._current_time_nanos
                                   - self._start_time_nanos))
            if self._current_fes > 0:
                self._log_best(kv)

        with logger.key_values(SECTION_SETUP) as kv:
            self.log_parameters_to(kv)

        log_sys_info(logger)

        if self._current_fes > 0:
            self._write_result(logger)

    def _validate_x(self) -> None:
        """Validate x, if it exists."""

    def _check_timing(self) -> None:
        """
        Check whether there has been any timing errors.

        :raises ValueError: if there is any timing error
        """
        if self._current_time_nanos < self._start_time_nanos:
            raise ValueError(
                f"current_time_nanos={self._current_time_nanos} < "
                f"start_time_nanos={self._start_time_nanos}")
        if self._current_fes <= 0:
            raise ValueError("no FE was performed")
        if self._current_fes < self._last_improvement_fe:
            raise ValueError(
                f"current_fe={self._current_fes} < "
                f"last_improvement_fe={self._last_improvement_fe}")
        if self._current_time_nanos < self._last_improvement_time_nanos:
            raise ValueError(
                f"current_time_nanos={self._current_time_nanos} < "
                "last_improvement_time_nanos="
                f"{self._last_improvement_time_nanos}")

    def _validate_best_f(self) -> None:
        """
        Validate the best encountered objective value.

        :raises ValueError: if there is an error
        """
        ff: Final[int | float] = self._f(self._current_best_y)
        if ff != self._current_best_f:
            raise ValueError(  # noqa
                "We re-computed the objective value of the best solution"
                f" and got {ff}, but it has been registered as "
                f"{self._current_best_f}!")  # noqa
        if not isfinite(ff):
            raise ValueError(  # noqa
                f"Reproduced the objective value {ff} of the best "
                "solution, but it is not finite?")  # noqa

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        """Exit the process and write the log if necessary."""
        # noinspection PyProtectedMember
        super().__exit__(exception_type, exception_value, traceback)

        # Update the total consumed time, but not include the error checks
        # below.
        self._current_time_nanos = _TIME_IN_NS()

        y_error: BaseException | None = None  # error in solution?
        v_error: BaseException | None = None  # error in objective value?
        x_error: BaseException | None = None  # error in search space?
        t_error: BaseException | None = None  # error in timing?
        log_error: BaseException | None = None  # error while logging?
        try:
            self._solution_space.validate(self._current_best_y)
        except BaseException as be:
            y_error = be
        if self._current_fes > 0:
            try:
                self._validate_best_f()
            except BaseException as be:
                v_error = be
            try:
                self._validate_x()
            except BaseException as be:
                x_error = be
        try:
            self._check_timing()
        except BaseException as be:
            t_error = be

        if self.__log_file is not None:
            with FileLogger(self.__log_file) as logger:
                try:
                    self._write_log(logger)
                except BaseException as be:
                    log_error = be

                if self._caught is not None:
                    _error_2(logger, SECTION_ERROR_IN_RUN,
                             self._caught)
                if exception_type or exception_value or traceback:
                    _error_1(logger, SECTION_ERROR_IN_CONTEXT,
                             exception_type, exception_value, traceback)
                if y_error:
                    _error_2(logger, SECTION_ERROR_INVALID_Y, y_error)
                if v_error:
                    _error_2(logger, SECTION_ERROR_BEST_F, v_error)
                if x_error:
                    _error_2(logger, SECTION_ERROR_INVALID_X, x_error)
                if t_error:
                    _error_2(logger, SECTION_ERROR_TIMING, t_error)
                if log_error:
                    _error_2(logger, SECTION_ERROR_IN_LOG, log_error)

                # flush all the additional log sections at the end
                for t in sorted(self.__sections.keys()):
                    with logger.text(t) as sec:
                        sec.write(self.__sections[t])
                del self.__sections

        if not exception_type:
            # if no error happened when closing the process, raise any error
            # caught during validation.
            if self._caught is not None:
                raise self._caught  # pylint: disable=[E0702]
            if y_error:
                raise y_error
            if v_error:
                raise v_error
            if x_error:
                raise x_error
            if t_error:
                raise t_error
            if log_error:
                raise log_error

    def __str__(self) -> str:
        """
        Get the name of this process implementation.

        :return: "baseProcess"
        """
        return "baseProcess"


def _check_log_time(start_time: int, current_time: int,
                    log: list[list]) -> None:
    """
    Check the time inside the log.

    :param start_time: the start time
    :param current_time: the current time
    :param log: the log
    :raises ValueError: if there is a timing error in the log
    """
    last_time: int = -1
    last_fe: int = -1
    for row in log:
        fes: int = cast(int, row[0])
        time: int = cast(int, row[1])
        if fes < last_fe:
            raise ValueError(f"fe={fes} after fe={last_fe}?")
        if time < last_time:
            raise ValueError(
                f"time={time} of fe={fes} is less than "
                f"last_time={last_time} of last_fe={last_fe}")
        if time < start_time:
            raise ValueError(
                f"time={time} of fe={fes} is less than "
                f"start_time_nanos={start_time}")
        if time > current_time:
            raise ValueError(
                f"time={time} of fe={fes} is greater than "
                f"current_time_nanos={current_time}")
        last_time = time
        last_fe = fes



"""The algorithm execution API."""
from math import isfinite
from typing import Any, Final

from moptipy.api._process_base import _ProcessBase
from moptipy.api._process_no_ss import _ProcessNoSS
from moptipy.api._process_no_ss_log import _ProcessNoSSLog
from moptipy.api._process_ss import _ProcessSS
from moptipy.api._process_ss_log import _ProcessSSLog
from moptipy.api.algorithm import Algorithm, check_algorithm
from moptipy.api.encoding import Encoding, check_encoding
from moptipy.api.objective import Objective, check_objective
from moptipy.api.process import (
    Process,
    check_goal_f,
    check_max_fes,
    check_max_time_millis,
)
from moptipy.api.space import Space, check_space
from moptipy.utils.nputils import rand_seed_check
from moptipy.utils.path import Path
from moptipy.utils.types import type_error


def _check_log_file(log_file: Any, none_is_ok: bool = True) -> Path | None:
    """
    Check a log file.

    :param log_file: the log file
    :param none_is_ok: is `None` ok for log files?
    :return: the log file

    >>> print(_check_log_file("/a/b.txt"))
    /a/b.txt
    >>> print(_check_log_file("/a/b.txt", False))
    /a/b.txt
    >>> print(_check_log_file("/a/b.txt", True))
    /a/b.txt
    >>> from moptipy.utils.path import Path as Pth
    >>> print(_check_log_file(Pth.path("/a/b.txt"), False))
    /a/b.txt
    >>> print(_check_log_file(None))
    None
    >>> print(_check_log_file(None, True))
    None
    >>> try:
    ...     _check_log_file(1)  # noqa  # type: ignore
    ... except TypeError as te:
    ...     print(te)
    path should be an instance of str but is int, namely '1'.
    >>> try:
    ...     _check_log_file(None, False)
    ... except TypeError as te:
    ...     print(te)
    path should be an instance of str but is None.
    """
    if (log_file is None) and none_is_ok:
        return None
    return Path.path(log_file)


class Execution:
    """
    Define all the components of an experiment and then execute it.

    This class follows the builder pattern. It allows us to
    step-by-step store all the parameters needed to execute an
    experiment. Via the method :meth:`~Execution.execute`, we can then
    run the experiment and obtain the instance of
    :class:`~moptipy.api.process.Process` *after* the execution of the
    algorithm. From this instance, we can query the final result of the
    algorithm application.
    """

    def __init__(self) -> None:
        """Initialize the execution builder."""
        super().__init__()
        self._algorithm: Algorithm | None = None
        self._solution_space: Space | None = None
        self._objective: Objective | None = None
        self._search_space: Space | None = None
        self._encoding: Encoding | None = None
        self._rand_seed: int | None = None
        self._max_fes: int | None = None
        self._max_time_millis: int | None = None
        self._goal_f: None | int | float = None
        self._log_file: Path | None = None
        self._log_improvements: bool = False
        self._log_all_fes: bool = False

    def set_algorithm(self, algorithm: Algorithm) -> "Execution":
        """
        Set the algorithm to be used for this experiment.

        :param algorithm: the algorithm
        :returns: this execution
        """
        self._algorithm = check_algorithm(algorithm)
        return self

    def set_solution_space(self, solution_space: Space) -> "Execution":
        """
        Set the solution space to be used for this experiment.

        This is the space managing the data structure holding the candidate
        solutions.

        :param solution_space: the solution space
        :returns: this execution
        """
        self._solution_space = check_space(solution_space)
        return self

    def set_objective(self, objective: Objective) -> "Execution":
        """
        Set the objective function to be used for this experiment.

        This is the function rating the quality of candidate solutions.

        :param objective: the objective function
        :returns: this execution
        """
        if self._objective is not None:
            raise ValueError(
                "Cannot add more than one objective function in single-"
                f"objective optimization, attempted to add {objective} "
                f"after {self._objective}.")
        self._objective = check_objective(objective)
        return self

    def set_search_space(self, search_space: Space | None) -> "Execution":
        """
        Set the search space to be used for this experiment.

        This is the space from which the algorithm samples points.

        :param search_space: the search space, or `None` of none shall be
            used, i.e., if search and solution space are the same
        :returns: this execution
        """
        self._search_space = check_space(search_space, none_is_ok=True)
        return self

    def set_encoding(self, encoding: Encoding | None) -> "Execution":
        """
        Set the encoding to be used for this experiment.

        This is the function translating from the search space to the
        solution space.

        :param encoding: the encoding, or `None` of none shall be used
        :returns: this execution
        """
        self._encoding = check_encoding(encoding, none_is_ok=True)
        return self

    def set_rand_seed(self, rand_seed: int | None) -> "Execution":
        """
        Set the seed to be used for initializing the random number generator.

        :param rand_seed: the random seed, or `None` if a seed should
            automatically be chosen when the experiment is executed
        """
        self._rand_seed = None if rand_seed is None \
            else rand_seed_check(rand_seed)
        return self

    def set_max_fes(self, max_fes: int,  # +book
                    force_override: bool = False) -> "Execution":
        """
        Set the maximum FEs.

        This is the number of candidate solutions an optimization is allowed
        to evaluate. If this method is called multiple times, then the
        shortest limit is used unless `force_override` is `True`.

        :param max_fes: the maximum FEs
        :param force_override: the use the value given in `max_time_millis`
            regardless of what was specified before
        :returns: this execution
        """
        max_fes = check_max_fes(max_fes)
        if (self._max_fes is not None) and (max_fes >= self._max_fes) \
                and (not force_override):
            return self
        self._max_fes = max_fes
        return self

    def set_max_time_millis(self, max_time_millis: int,
                            force_override: bool = False) -> "Execution":
        """
        Set the maximum time in milliseconds.

        This is the maximum time that the process is allowed to run. If this
        method is called multiple times, the shortest time is used unless
        `force_override` is `True`.

        :param max_time_millis: the maximum time in milliseconds
        :param force_override: the use the value given in `max_time_millis`
            regardless of what was specified before
        :returns: this execution
        """
        max_time_millis = check_max_time_millis(max_time_millis)
        if (self._max_time_millis is not None) \
                and (max_time_millis >= self._max_time_millis) \
                and (not force_override):
            return self
        self._max_time_millis = max_time_millis
        return self

    def set_goal_f(self, goal_f: int | float) -> "Execution":
        """
        Set the goal objective value after which the process can stop.

        If this method is called multiple times, then the largest value is
        retained.

        :param goal_f: the goal objective value.
        :returns: this execution
        """
        goal_f = check_goal_f(goal_f)
        if (self._goal_f is not None) and (goal_f <= self._goal_f):
            return self
        self._goal_f = goal_f
        return self

    def set_log_file(self, log_file: str | None) -> "Execution":
        """
        Set the log file to write to.

        If a path to a log file is provided, the contents of this file will be
        filled based on the structure documented at
        https://thomasweise.github.io/moptipy/#log-file-sections, which
        includes the algorithm parameters, the instance features, the
        system settings, the final solution, the corresponding point in the
        search space, etc.

        This method can be called arbitrarily often.

        :param log_file: the log file
        """
        self._log_file = _check_log_file(log_file, True)
        return self

    def set_log_improvements(self, log_improvements: bool = True) \
            -> "Execution":
        """
        Set whether improvements should be logged.

        If improvements are logged, then the `PROGRESS` section will be added
        to the log files, as documented at
        https://thomasweise.github.io/moptipy/#the-section-progress.

        :param log_improvements: if improvements should be logged?
        :returns: this execution
        """
        if not isinstance(log_improvements, bool):
            raise type_error(log_improvements, "log_improvements", bool)
        self._log_improvements = log_improvements
        return self

    def set_log_all_fes(self, log_all_fes: bool = True) -> "Execution":
        """
        Set whether all objective function evaluations (FEs) should be logged.

        If all FEs  are logged, then the `PROGRESS` section will be added to
        the log files, as documented at
        https://thomasweise.github.io/moptipy/#the-section-progress.

        :param log_all_fes: if all FEs should be logged?
        :returns: this execution
        """
        if not isinstance(log_all_fes, bool):
            raise type_error(log_all_fes, "log_all_fes", bool)
        self._log_all_fes = log_all_fes
        return self

    def execute(self) -> Process:
        """
        Execute the experiment and return the process *after* the run.

        The optimization process constructed with this object is executed.
        This means that first, an instance of
        :class:`~moptipy.api.process.Process` is constructed.
        Then, the method :meth:`~moptipy.api.algorithm.Algorithm.solve` is
        applied to this instance.
        In other words, the optimization algorithm is executed until it
        terminates.
        Finally, this method returns the :class:`~moptipy.api.process.Process`
        instance *after* algorithm completion.
        This instance then can be queried for the final result of the run (via
        :meth:`~moptipy.api.process.Process.get_copy_of_best_y`), the
        objective value of this final best solution (via
        :meth:`~moptipy.api.process.Process.get_best_f`), and other
        information.

        If a log file path was supplied to :meth:`~set_log_file`, then the
        information gathered from the optimization process will be written to
        the file *after*  the `with` blog using the process is left. See
        https://thomasweise.github.io/moptipy/#log-file-sections for a
        documentation of the log file structure and sections.

        :return: the process *after* the run, i.e., in the state where it can
            be queried for the result
        """
        algorithm: Final[Algorithm] = check_algorithm(self._algorithm)
        solution_space: Final[Space] = check_space(self._solution_space)
        objective: Final[Objective] = check_objective(self._objective)
        search_space: Final[Space | None] = check_space(
            self._search_space, self._encoding is None)
        encoding: Final[Encoding | None] = check_encoding(
            self._encoding, search_space is None)
        rand_seed = self._rand_seed
        if rand_seed is not None:
            rand_seed = rand_seed_check(rand_seed)
        max_time_millis = check_max_time_millis(self._max_time_millis, True)
        max_fes = check_max_fes(self._max_fes, True)
        goal_f = check_goal_f(self._goal_f, True)
        f_lb = objective.lower_bound()
        if (f_lb is not None) and isfinite(f_lb) \
                and ((goal_f is None) or (f_lb > goal_f)):
            goal_f = f_lb

        log_all_fes = self._log_all_fes
        log_improvements = self._log_improvements or self._log_all_fes

        log_file = self._log_file
        if log_file is None:
            if log_all_fes:
                raise ValueError("Log file cannot be None "
                                 "if all FEs should be logged.")
            if log_improvements:
                raise ValueError("Log file cannot be None "
                                 "if improvements should be logged.")
        else:
            log_file.create_file_or_truncate()

        process: Final[_ProcessBase] = \
            (_ProcessNoSSLog(solution_space=solution_space,
                             objective=objective,
                             algorithm=algorithm,
                             log_file=log_file,
                             rand_seed=rand_seed,
                             max_fes=max_fes,
                             max_time_millis=max_time_millis,
                             goal_f=goal_f,
                             log_all_fes=log_all_fes)
             if log_improvements or log_all_fes else
             _ProcessNoSS(solution_space=solution_space,
                          objective=objective,
                          algorithm=algorithm,
                          log_file=log_file,
                          rand_seed=rand_seed,
                          max_fes=max_fes,
                          max_time_millis=max_time_millis,
                          goal_f=goal_f)) if search_space is None else \
            (_ProcessSSLog(solution_space=solution_space,
                           objective=objective,
                           algorithm=algorithm,
                           search_space=search_space,
                           encoding=encoding,
                           log_file=log_file,
                           rand_seed=rand_seed,
                           max_fes=max_fes,
                           max_time_millis=max_time_millis,
                           goal_f=goal_f,
                           log_all_fes=log_all_fes)
             if log_improvements or log_all_fes else
             _ProcessSS(solution_space=solution_space,
                        objective=objective,
                        algorithm=algorithm,
                        search_space=search_space,
                        encoding=encoding,
                        log_file=log_file,
                        rand_seed=rand_seed,
                        max_fes=max_fes,
                        max_time_millis=max_time_millis,
                        goal_f=goal_f))
        try:
            # noinspection PyProtectedMember
            process._after_init()  # finalize the created process
            objective.initialize()  # initialize the objective function
            if encoding is not None:
                encoding.initialize()   # initialize the encoding
            solution_space.initialize()  # initialize the solution space
            if search_space is not None:
                search_space.initialize()  # initialize the search space
            algorithm.initialize()  # initialize the algorithm
            algorithm.solve(process)  # apply the algorithm
        except BaseException as be:
            # noinspection PyProtectedMember
            process._caught = be
        return process



"""
A nullary operator shuffling a permutation.

This operator first copies the canonical permutation to the destination
string. It then applies the shuffle procedure of the random number
generator, which probably internally applies a Fisher-Yates Shuffle.
The result is a random permutation.

1. Thomas Weise. *Optimization Algorithms.* 2021. Hefei, Anhui, China:
   Institute of Applied Optimization (IAO), School of Artificial Intelligence
   and Big Data, Hefei University. http://thomasweise.github.io/oa/
"""
from typing import Final

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op0
from moptipy.spaces.permutations import Permutations
from moptipy.utils.types import type_error


# start book
class Op0Shuffle(Op0):
    """Shuffle permutations randomly."""

    def __init__(self, space: Permutations):
        """
        Initialize this shuffle operation: use blueprint from space.

        :param space: the search space
        """
        super().__init__()  # -book
        if not isinstance(space, Permutations):  # -book
            raise type_error(space, "space", Permutations)  # -book
        #: the internal blueprint for filling permutations
        self.__blueprint: Final[np.ndarray] = space.blueprint

    def op0(self, random: Generator, dest: np.ndarray) -> None:
        """
        Copy the base string to `dest` and shuffle it randomly.

        :param random: the random number generator
        :param dest: the permutation that should be shuffled.
            Afterwards, the order of its elements is random.
        """
        np.copyto(dest, self.__blueprint)  # Copy blueprint to dest.
        random.shuffle(dest)  # Shuffle destination array randomly.
    # end book

    def __str__(self) -> str:
        """
        Get the name of this operator.

        :return: "shuffle"
        """
        return "shuffle"



"""The base class for spaces based on numpy arrays."""
from typing import Final

import numpy as np

from moptipy.api.logging import KEY_SPACE_NUM_VARS
from moptipy.api.space import Space
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection
from moptipy.utils.nputils import (
    KEY_NUMPY_TYPE,
    array_to_str,
    numpy_type_to_str,
)
from moptipy.utils.types import check_int_range, type_error


class NPArraySpace(Space):
    """
    A space where each element is a one-dimensional :class:`numpy.ndarray`.

    Such spaces can serve as basis for implementing combinatorial
    optimization and can be extended to host permutations.

    >>> import numpy as npx
    >>> s = NPArraySpace(9, npx.dtype(int))
    >>> print(s.dimension)
    9
    >>> print(s.dtype)
    int64
    >>> print(s.create())
    [0 0 0 0 0 0 0 0 0]
    >>> print(s.to_str(s.create()))
    0;0;0;0;0;0;0;0;0
    >>> print(s.from_str(s.to_str(s.create())))
    [0 0 0 0 0 0 0 0 0]
    """

    def __init__(self, dimension: int, dtype: np.dtype) -> None:
        """
        Create the numpy array-based search space.

        :param dimension: The dimension of the search space,
            i.e., the number of decision variables.
        :param dtype: the data type
        """
        if not isinstance(dtype, np.dtype):
            raise type_error(dtype, "dtype", np.dtype)
        if (not isinstance(dtype.char, str)) or (len(dtype.char) != 1):
            raise ValueError(
                f"dtype.char must be str of length 1, but is {dtype.char}")
        #: The basic data type of the vector space elements.
        self.dtype: Final[np.dtype] = dtype
        #: The dimension, i.e., the number of elements of the vectors.
        self.dimension: Final[int] = check_int_range(
            dimension, "dimension", 1, 100_000_000)
        # the function forwards
        self.copy = np.copyto  # type: ignore
        self.is_equal = np.array_equal  # type: ignore
        self.to_str = array_to_str  # type: ignore

    def create(self) -> np.ndarray:
        """
        Create a vector with all zeros.

        :return: the vector
        """
        return np.zeros(shape=self.dimension, dtype=self.dtype)

    def from_str(self, text: str) -> np.ndarray:
        """
        Convert a string to a vector.

        :param text: the text
        :return: the vector
        :raises TypeError: if `text` is not a `str`
        :raises ValueError: if `text` cannot be converted to a valid vector
        """
        if not (isinstance(text, str)):
            raise type_error(text, "text", str)
        x = np.fromstring(text, dtype=self.dtype, sep=CSV_SEPARATOR)
        self.validate(x)
        return x

    def validate(self, x: np.ndarray) -> None:
        """
        Validate a numpy nd-array.

        :param x: the numpy vector
        :raises TypeError: if the string is not an element of this space.
        :raises ValueError: if the shape of the vector is wrong or any of its
            element is not finite.
        """
        if not isinstance(x, np.ndarray):
            raise type_error(x, "x", np.ndarray)
        if x.dtype != self.dtype:
            raise ValueError(
                f"x must be of type {self.dtype} but is of type {x.dtype}.")
        if (len(x.shape) != 1) or (x.shape[0] != self.dimension):
            raise ValueError(f"x must be of shape ({self.dimension}) but is "
                             f"of shape {x.shape}.")

    def __str__(self) -> str:
        """
        Get the name of this np array space.

        :return: "ndarray" + dimension + dtype.char

        >>> import numpy as npx
        >>> print(NPArraySpace(4, npx.dtype(int)))
        ndarray4l
        """
        return f"ndarray{self.dimension}{self.dtype.char}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this space to the given logger.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> import numpy as npx
        >>> dt = npx.dtype(float)
        >>> dt.char
        'd'
        >>> space = NPArraySpace(10, dt)
        >>> space.dimension
        10
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         space.log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[-2]
        'dtype: d'
        >>> text[-3]
        'nvars: 10'
        >>> text[-4]
        'class: moptipy.spaces.nparrayspace.NPArraySpace'
        >>> text[-5]
        'name: ndarray10d'
        >>> len(text)
        6
        """
        super().log_parameters_to(logger)
        logger.key_value(KEY_SPACE_NUM_VARS, self.dimension)
        logger.key_value(KEY_NUMPY_TYPE, numpy_type_to_str(self.dtype))
